import datetime
import os
import random
import shutil
from pathlib import Path
from typing import Any, Dict

import numpy as np
import polars as pl
import torch
import torch.distributed as dist
from cehrbert.runners.runner_util import generate_prepared_ds_path
from datasets import load_from_disk
from meds import held_out_split, train_split, tuning_split
from torch.utils.data import DataLoader
from tqdm import tqdm
from transformers.trainer_utils import is_main_process
from transformers.utils import is_flash_attn_2_available, logging

from cehrgpt.data.hf_cehrgpt_dataset import create_cehrgpt_finetuning_dataset
from cehrgpt.data.hf_cehrgpt_dataset_collator import CehrGptDataCollator
from cehrgpt.generation.generate_batch_hf_gpt_sequence import (
    generate_single_batch,
    normalize_value,
)
from cehrgpt.gpt_utils import (
    extract_time_interval_in_days,
    extract_time_interval_in_hours,
    is_att_token,
    is_inpatient_hour_token,
    is_visit_end,
    is_visit_start,
)
from cehrgpt.models.hf_cehrgpt import CEHRGPT2LMHeadModel
from cehrgpt.models.tokenization_hf_cehrgpt import CehrGptTokenizer
from cehrgpt.runners.data_utils import (
    extract_cohort_sequences,
    prepare_finetune_dataset,
)
from cehrgpt.runners.gpt_runner_util import parse_runner_args
from cehrgpt.runners.hf_cehrgpt_pretrain_runner import tokenizer_exists

LOG = logging.get_logger("transformers")


def map_data_split_name(split: str) -> str:
    if split == "train":
        return train_split
    elif split == "validation":
        return tuning_split
    elif split == "test":
        return held_out_split
    raise ValueError(f"Unknown split: {split}")


def seed_all(seed: int = 42):
    """Set seed for Python, NumPy, and PyTorch (CPU & CUDA)."""
    random.seed(seed)  # Python random
    np.random.seed(seed)  # NumPy
    torch.manual_seed(seed)  # PyTorch CPU
    torch.cuda.manual_seed(seed)  # Current GPU
    torch.cuda.manual_seed_all(seed)  # All GPUs

    # For reproducibility in dataloader workers
    os.environ["PYTHONHASHSEED"] = str(seed)


def generate_trajectories_per_batch(
    batch: Dict[str, Any],
    cehrgpt_tokenizer: CehrGptTokenizer,
    cehrgpt_model: CEHRGPT2LMHeadModel,
    device,
    data_output_path: Path,
    max_length: int,
):
    subject_ids = batch["person_id"].squeeze().detach().cpu().tolist()
    prediction_times = batch["index_date"].squeeze().detach().cpu().tolist()
    batched_epoch_times = batch["epoch_times"].detach().cpu().tolist()
    batched_input_ids = batch["input_ids"]
    batched_ages = batch["ages"]
    batched_value_indicators = batch["value_indicators"]
    batched_values = batch["values"]
    # Make sure the batch does not exceed batch_size
    batch_sequences = generate_single_batch(
        cehrgpt_model,
        cehrgpt_tokenizer,
        batched_input_ids,
        ages=batched_ages,
        values=batched_values,
        value_indicators=batched_value_indicators,
        max_length=max_length,
        top_p=1.0,
        top_k=cehrgpt_tokenizer.vocab_size,
        device=device,
    )
    # Clear the cache
    torch.cuda.empty_cache()

    trajectories = []
    for sample_i, (concept_ids, value_indicators, values) in enumerate(
        zip(
            batch_sequences["sequences"],
            batch_sequences["value_indicators"],
            batch_sequences["values"],
        )
    ):
        (
            concept_ids,
            is_numeric_types,
            number_as_values,
            concept_as_values,
            units,
        ) = normalize_value(concept_ids, values, cehrgpt_tokenizer)

        epoch_times = batched_epoch_times[sample_i]
        input_length = len(epoch_times)
        # Getting the last observed event time from the token before the prediction time
        window_last_observed = epoch_times[input_length - 1]
        current_cursor = epoch_times[-1]
        generated_epoch_times = []
        valid_indices = []

        for i in range(input_length, len(concept_ids)):
            concept_id = concept_ids[i]
            # We use the left padding strategy in the data collator
            if concept_id in [cehrgpt_tokenizer.pad_token, cehrgpt_tokenizer.end_token]:
                continue
            # We need to construct the time stamp
            if is_att_token(concept_id):
                current_cursor += extract_time_interval_in_days(concept_id) * 24 * 3600
            elif is_inpatient_hour_token(concept_id):
                current_cursor += extract_time_interval_in_hours(concept_id) * 3600
            elif is_visit_start(concept_id) or is_visit_end(concept_id):
                continue
            else:
                valid_indices.append(i)
                generated_epoch_times.append(
                    datetime.datetime.utcfromtimestamp(current_cursor).replace(
                        tzinfo=None
                    )
                )

        trajectories.append(
            {
                "subject_id": subject_ids[sample_i],
                "prediction_time": datetime.datetime.utcfromtimestamp(
                    prediction_times[sample_i]
                ).replace(tzinfo=None),
                "window_last_observed_time": datetime.datetime.utcfromtimestamp(
                    window_last_observed
                ).replace(tzinfo=None),
                "times": generated_epoch_times,
                "concept_ids": np.asarray(concept_ids)[valid_indices].tolist(),
                "numeric_values": np.asarray(number_as_values)[valid_indices].tolist(),
                "text_value": np.asarray(concept_as_values)[valid_indices].tolist(),
                "units": np.asarray(units)[valid_indices].tolist(),
            }
        )

    trajectories = (
        pl.DataFrame(trajectories)
        .explode(["times", "concept_ids", "numeric_values", "text_value", "units"])
        .rename(
            {
                "times": "time",
                "concept_ids": "code",
                "numeric_values": "numeric_value",
                "units": "unit",
            }
        )
        .select(
            "subject_id",
            "prediction_time",
            "window_last_observed_time",
            "time",
            "code",
            "numeric_value",
            "text_value",
            "unit",
        )
    )
    trajectories.write_parquet(data_output_path)


def main():
    cehrgpt_args, data_args, model_args, training_args = parse_runner_args()
    if torch.cuda.is_available():
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")
    cehrgpt_tokenizer = CehrGptTokenizer.from_pretrained(
        model_args.tokenizer_name_or_path
    )
    cehrgpt_model = (
        CEHRGPT2LMHeadModel.from_pretrained(
            model_args.model_name_or_path,
            attn_implementation=(
                "flash_attention_2" if is_flash_attn_2_available() else "eager"
            ),
        )
        .eval()
        .to(device)
    )
    cehrgpt_model.generation_config.pad_token_id = cehrgpt_tokenizer.pad_token_id
    cehrgpt_model.generation_config.eos_token_id = cehrgpt_tokenizer.end_token_id
    cehrgpt_model.generation_config.bos_token_id = cehrgpt_tokenizer.end_token_id

    if not os.path.exists(training_args.output_dir):
        os.makedirs(training_args.output_dir)

    prepared_ds_path = generate_prepared_ds_path(
        data_args, model_args, data_folder=data_args.cohort_folder
    )

    processed_dataset = None
    if any(prepared_ds_path.glob("*")):
        LOG.info(f"Loading prepared dataset from disk at {prepared_ds_path}...")
        processed_dataset = load_from_disk(str(prepared_ds_path))
        LOG.info("Prepared dataset loaded from disk...")
        if cehrgpt_args.expand_tokenizer:
            if tokenizer_exists(training_args.output_dir):
                cehrgpt_tokenizer = CehrGptTokenizer.from_pretrained(
                    training_args.output_dir
                )
            else:
                LOG.warning(
                    f"CehrGptTokenizer must exist in {training_args.output_dir} "
                    f"when the dataset has been processed and expand_tokenizer is set to True. "
                    f"Please delete the processed dataset at {prepared_ds_path}."
                )
                processed_dataset = None
                shutil.rmtree(prepared_ds_path)

    if processed_dataset is None and is_main_process(training_args.local_rank):
        # If the full dataset has been tokenized, we don't want to tokenize the cohort containing
        # the subset of the data. We should slice out the portion of the tokenized sequences for each sample
        if cehrgpt_args.tokenized_full_dataset_path is not None:
            processed_dataset = extract_cohort_sequences(data_args, cehrgpt_args)
        else:
            # Organize them into a single DatasetDict
            final_splits = prepare_finetune_dataset(
                data_args, training_args, cehrgpt_args
            )
            # TODO: temp solution, this column is mixed typed and causes an issue when transforming the data
            if not data_args.streaming:
                all_columns = final_splits["train"].column_names
                if "visit_concept_ids" in all_columns:
                    final_splits = final_splits.remove_columns(["visit_concept_ids"])

            processed_dataset = create_cehrgpt_finetuning_dataset(
                dataset=final_splits,
                cehrgpt_tokenizer=cehrgpt_tokenizer,
                data_args=data_args,
            )
        if not data_args.streaming:
            processed_dataset.save_to_disk(prepared_ds_path)
            processed_dataset.cleanup_cache_files()

    # After main-process-only operations, synchronize all processes to ensure consistency
    if dist.is_available() and dist.is_initialized():
        dist.barrier()

    # We suppress the additional learning objectives in fine-tuning
    data_collator = CehrGptDataCollator(
        tokenizer=cehrgpt_tokenizer,
        max_length=cehrgpt_args.generation_input_length,
        include_values=cehrgpt_model.config.include_values,
        pretraining=False,
        include_ttv_prediction=False,
        use_sub_time_tokenization=False,
        include_demographics=False,
        add_linear_prob_token=False,
    )

    LOG.info(
        "Generating %s trajectories per sample",
        cehrgpt_args.num_of_trajectories_per_sample,
    )
    for sample_i in range(cehrgpt_args.num_of_trajectories_per_sample):
        for split, dataset in processed_dataset.items():
            meds_split = map_data_split_name(split)
            dataloader = DataLoader(
                dataset=dataset,
                batch_size=training_args.per_device_eval_batch_size,
                num_workers=training_args.dataloader_num_workers,
                collate_fn=data_collator,
                pin_memory=training_args.dataloader_pin_memory,
            )
            sample_output_dir = (
                Path(training_args.output_dir) / meds_split / f"{sample_i}"
            )
            sample_output_dir.mkdir(exist_ok=True, parents=True)
            for batch_i, batch in tqdm(
                enumerate(dataloader),
                desc=f"Generating Trajectories for split {meds_split} with trajectory {sample_i + 1}",
            ):
                output_parquet_file = sample_output_dir / f"{batch_i}.parquet"
                if output_parquet_file.exists():
                    LOG.info("%s already exists, skip...", output_parquet_file)
                    continue

                generate_trajectories_per_batch(
                    batch,
                    cehrgpt_tokenizer,
                    cehrgpt_model,
                    device,
                    sample_output_dir / f"{batch_i}.parquet",
                    cehrgpt_args.generation_max_new_tokens
                    + cehrgpt_args.generation_input_length,
                )


if __name__ == "__main__":
    # ✅ Call first thing inside main()
    seed_all(42)
    main()



import argparse
import configparser
from pathlib import Path

import pyspark.sql.functions as f
from pyspark.sql import SparkSession

# Define timestamp column for filtering based on the folder name
omop_timestamp_dict = {
    "person": "birth_datetime",
    "condition_occurrence": "condition_start_date",
    "measurement": "measurement_date",
    "drug_exposure": "drug_exposure_start_date",
    "procedure_occurrence": "procedure_date",
    "observation": "observation_date",
    "visit_occurrence": "visit_start_date",
}


# Function to initialize and return the SparkSession
def get_spark_session():
    spark = (
        SparkSession.builder.appName("OMOP Upload")
        .config("spark.sql.legacy.parquet.int96RebaseModeInRead", "CORRECTED")
        .config("spark.sql.legacy.parquet.int96RebaseModeInWrite", "CORRECTED")
        .config("spark.sql.legacy.parquet.datetimeRebaseModeInRead", "CORRECTED")
        .config("spark.sql.legacy.parquet.datetimeRebaseModeInWrite", "CORRECTED")
        .getOrCreate()
    )
    return spark


# Function to upload OMOP tables to a database
def upload_omop_tables(spark, domain_table_folder, db_properties):
    # Load parquet file from the specified folder
    df = spark.read.format("parquet").load(str(domain_table_folder) + "/")

    # Filter dates outside of the acceptable range
    if domain_table_folder.name in omop_timestamp_dict:
        timestamp_column = omop_timestamp_dict[domain_table_folder.name]
        df = df.filter(f.col(timestamp_column) > f.lit("1900-01-01").cast("date"))
        df = df.filter(f.col(timestamp_column) < f.lit("9999-01-01").cast("date"))

    # Cast appropriate columns to integer and date types
    for column in df.columns:
        if "concept_id" in column:
            df = df.withColumn(column, f.col(column).cast("integer"))
        if "date" in column:
            df = df.withColumn(column, f.col(column).cast("date"))

    # Write to the database with specified options
    df.repartition(10).write.format("jdbc").options(
        url=db_properties["base_url"],
        dbtable=domain_table_folder.name,
        user=db_properties["user"],
        password=db_properties["password"],
        batchsize=200000,
        queryTimeout=500,
    ).mode("overwrite").save()


# Main function to process the folders and upload tables
def main(credential_path, input_folder):
    # Load database properties from the credentials file
    config = configparser.ConfigParser()
    config.read(credential_path)
    db_properties = config.defaults()

    # Initialize SparkSession
    spark = get_spark_session()

    # Process each folder within the input folder
    input_folder = Path(input_folder)
    uploaded_tables = []
    for folder in input_folder.glob("*"):
        try:
            if folder.is_dir() and folder.name in omop_timestamp_dict:
                upload_omop_tables(spark, folder, db_properties)
                uploaded_tables.append(folder.name)
                print(f"Table: {folder.name} uploaded successfully")
        except Exception as e:
            print(f"Error uploading table {folder.name}: {e}")

    print(f"Uploaded tables: {uploaded_tables}")


# Argument parsing moved under __name__ == "__main__"
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Arguments for uploading OMOP tables")

    parser.add_argument(
        "-c",
        "--credential_path",
        required=True,
        help="The path for your database credentials",
    )

    parser.add_argument(
        "-i",
        "--input_folder",
        required=True,
        help="Path to the input folder containing the OMOP tables",
    )

    args = parser.parse_args()

    # Call the main function with parsed arguments
    main(args.credential_path, args.input_folder)



import glob
import os.path

from cehrgpt.omop.omop_argparse import create_omop_argparse
from cehrgpt.omop.omop_table_builder import OmopTableBuilder
from cehrgpt.omop.queries.observation_period import (
    OBSERVATION_PERIOD_QUERY,
    OBSERVATION_PERIOD_WITH_MEASUREMENT_QUERY,
)

OBSERVATION_PERIOD = "observation_period"


def main(args):
    include_measurement = (
        len(glob.glob(os.path.join(args.input_folder, "measurement", "*.parquet"))) > 0
    )
    dependency_list = [
        "person",
        "visit_occurrence",
        "condition_occurrence",
        "procedure_occurrence",
        "drug_exposure",
    ]
    if include_measurement:
        dependency_list.append("measurement")

    OmopTableBuilder.create_omop_query_builder(
        input_folder=args.input_folder,
        output_folder=args.output_folder,
        continue_job=args.continue_job,
        table_name=OBSERVATION_PERIOD,
        query_template=(
            OBSERVATION_PERIOD_WITH_MEASUREMENT_QUERY
            if include_measurement
            else OBSERVATION_PERIOD_QUERY
        ),
        dependency_list=dependency_list,
    ).build()


if __name__ == "__main__":
    main(create_omop_argparse().parse_args())



import argparse
import os

from pyspark.sql import SparkSession

# Define timestamp column for filtering based on the folder name
omop_tables = [
    "person",
    "death",
    "visit_occurrence",
    "condition_occurrence",
    "procedure_occurrence",
    "drug_exposure",
    "measurement",
    "observation",
    "observation_period",
    "condition_era",
    "drug_era",
]


# Main function to process the folders and upload tables
def main(args):
    spark = (
        SparkSession.builder.appName("Sample OMOP Tables")
        .config("spark.sql.legacy.parquet.int96RebaseModeInRead", "CORRECTED")
        .config("spark.sql.legacy.parquet.int96RebaseModeInWrite", "CORRECTED")
        .config("spark.sql.legacy.parquet.datetimeRebaseModeInRead", "CORRECTED")
        .config("spark.sql.legacy.parquet.datetimeRebaseModeInWrite", "CORRECTED")
        .getOrCreate()
    )
    patient_sample = spark.read.parquet(args.person_sample)
    for omop_table in omop_tables:
        omop_table_input = os.path.join(args.omop_folder, omop_table)
        omop_table_output = os.path.join(args.output_folder, omop_table)
        # If the input does not exist, let's skip
        if not os.path.exists(omop_table_input):
            continue
        # If the output already exists, and overwrite is not set to True, let's skip it
        if os.path.exists(omop_table_output) and not args.overwrite:
            continue
        omop_dataframe = spark.read.parquet(omop_table_input)
        sub_omop_dataframe = omop_dataframe.join(
            patient_sample.select("person_id"), "person_id"
        )
        sub_omop_dataframe.write.mode("overwrite" if args.overwrite else None).parquet(
            omop_table_output
        )


# Argument parsing moved under __name__ == "__main__"
if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Arguments for uploading OMOP tables")
    parser.add_argument(
        "--person_sample",
        required=True,
    )
    parser.add_argument(
        "--omop_folder",
        required=True,
    )
    parser.add_argument(
        "--output_folder",
        required=True,
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
    )
    # Call the main function with parsed arguments
    main(parser.parse_args())



import os
from functools import partial
from pathlib import Path
from typing import Optional, Union

import datasets
import numpy as np
import pandas as pd
import torch
import torch.distributed as dist
from cehrbert.data_generators.hf_data_generator.meds_utils import (
    CacheFileCollector,
    create_dataset_from_meds_reader,
)
from cehrbert.runners.hf_runner_argument_dataclass import (
    DataTrainingArguments,
    ModelArguments,
)
from cehrbert.runners.runner_util import (
    generate_prepared_ds_path,
    get_last_hf_checkpoint,
    get_meds_extension_path,
    load_parquet_as_dataset,
)
from datasets import Dataset, DatasetDict, IterableDatasetDict, load_from_disk
from transformers import EarlyStoppingCallback, Trainer, set_seed
from transformers.trainer_utils import is_main_process
from transformers.utils import is_flash_attn_2_available, logging

from cehrgpt.data.hf_cehrgpt_dataset import create_cehrgpt_pretraining_dataset
from cehrgpt.data.hf_cehrgpt_dataset_collator import (
    CehrGptDataCollator,
    SamplePackingCehrGptDataCollator,
)
from cehrgpt.data.hf_cehrgpt_dataset_mapping import MedToCehrGPTDatasetMapping
from cehrgpt.models.config import CEHRGPTConfig
from cehrgpt.models.hf_cehrgpt import CEHRGPT2LMHeadModel
from cehrgpt.models.pretrained_embeddings import PretrainedEmbeddings
from cehrgpt.models.tokenization_hf_cehrgpt import CehrGptTokenizer
from cehrgpt.omop.ontology import Ontology
from cehrgpt.runners.data_utils import get_torch_dtype
from cehrgpt.runners.gpt_runner_util import parse_runner_args
from cehrgpt.runners.hf_gpt_runner_argument_dataclass import CehrGPTArguments
from cehrgpt.runners.sample_packing_trainer import SamplePackingTrainer

LOG = logging.get_logger("transformers")


class CustomEarlyStoppingCallback(EarlyStoppingCallback):
    def check_metric_value(self, args, state, control, metric_value):
        # best_metric is set by code for load_best_model
        operator = np.greater if args.greater_is_better else np.less
        if state.best_metric is None or (
            operator(metric_value, state.best_metric)
            and abs(metric_value - state.best_metric) / state.best_metric
            > self.early_stopping_threshold
        ):
            self.early_stopping_patience_counter = 0
        else:
            self.early_stopping_patience_counter += 1


def tokenizer_exists(tokenizer_name_or_path: str) -> bool:
    # Try to load the pretrained tokenizer
    try:
        CehrGptTokenizer.from_pretrained(os.path.abspath(tokenizer_name_or_path))
        return True
    except Exception:
        LOG.info(f"The tokenizer does not exist at {tokenizer_name_or_path}")
        return False


def load_and_create_tokenizer(
    data_args: DataTrainingArguments,
    model_args: ModelArguments,
    cehrgpt_args: CehrGPTArguments,
    dataset: Union[Dataset, DatasetDict],
) -> CehrGptTokenizer:
    # Try to load the pretrained tokenizer
    tokenizer_abspath = os.path.expanduser(model_args.tokenizer_name_or_path)
    if not tokenizer_exists(tokenizer_abspath):
        if cehrgpt_args.include_motor_time_to_event and not cehrgpt_args.vocab_dir:
            raise RuntimeError(
                "motor_vocab_dir must be specified if include_motor_time_to_event is True"
            )
        ontology: Optional[Ontology] = None
        concept_name_mapping = {}
        if cehrgpt_args.vocab_dir:
            LOG.info("Loading concept data from disk at %s", cehrgpt_args.vocab_dir)
            concept_pd = pd.read_parquet(
                os.path.join(cehrgpt_args.vocab_dir, "concept")
            )
            for row in concept_pd.itertuples():
                concept_name_mapping[str(getattr(row, "concept_id"))] = getattr(
                    row, "concept_name"
                )

            if cehrgpt_args.motor_use_ontology:
                LOG.info("Creating ontology for MOTOR TTE predictions")
                ontology = Ontology(cehrgpt_args.vocab_dir)
                train_val_dataset = datasets.concatenate_datasets(
                    [dataset["train"], dataset["validation"]]
                )
                ontology.prune_to_dataset(
                    train_val_dataset,
                    num_proc=data_args.preprocessing_num_workers,
                    remove_ontologies={"SPL", "HemOnc", "LOINC"},
                )

        LOG.info("Started training the tokenizer ...")
        train_val_dataset = datasets.concatenate_datasets(
            [dataset["train"], dataset["validation"]]
        )
        tokenizer = CehrGptTokenizer.train_tokenizer(
            train_val_dataset,
            concept_name_mapping,
            data_args,
            PretrainedEmbeddings(cehrgpt_args.pretrained_embedding_path),
            num_motor_tasks=(
                cehrgpt_args.num_motor_tasks
                if cehrgpt_args.include_motor_time_to_event
                else None
            ),
            apply_entropy_filter=cehrgpt_args.apply_entropy_filter,
            min_prevalence=cehrgpt_args.min_prevalence,
            ontology=ontology,
        )
        LOG.info("Finished training the tokenizer ...")
        tokenizer.save_pretrained(tokenizer_abspath)
        LOG.info("Saved the tokenizer to %s", tokenizer_abspath)
    else:
        LOG.info("The tokenizer exists and will be loaded from %s", tokenizer_abspath)
        tokenizer = CehrGptTokenizer.from_pretrained(tokenizer_abspath)
    return tokenizer


def load_and_create_model(
    model_args: ModelArguments,
    cehrgpt_args: CehrGPTArguments,
    tokenizer: CehrGptTokenizer,
) -> CEHRGPT2LMHeadModel:
    attn_implementation = (
        "flash_attention_2" if is_flash_attn_2_available() else "eager"
    )
    torch_dtype = get_torch_dtype(model_args.torch_dtype)
    model_abspath = os.path.expanduser(model_args.model_name_or_path)
    if cehrgpt_args.continue_pretrain:
        try:
            pretrained_model = CEHRGPT2LMHeadModel.from_pretrained(
                model_abspath,
                attn_implementation=attn_implementation,
                torch_dtype=torch_dtype,
            )
            if (
                pretrained_model.config.max_position_embeddings
                < model_args.max_position_embeddings
            ):
                LOG.info(
                    f"Increase model.config.max_position_embeddings to {model_args.max_position_embeddings}"
                )
                pretrained_model.config.max_position_embeddings = (
                    model_args.max_position_embeddings
                )
                pretrained_model.resize_position_embeddings(
                    model_args.max_position_embeddings
                )
            return pretrained_model
        except Exception as e:
            LOG.error(
                f"When continue_pretrain is set to True, it assumes that CEHR-GPT has been trained "
                f"and will be used to pretrain on new datasets. The CEHR-GPT checkpoint must exist at {model_abspath}"
            )
            raise e
    try:
        model_config = CEHRGPTConfig.from_pretrained(
            model_abspath, attn_implementation=attn_implementation
        )
    except Exception as e:
        LOG.warning(e)
        if cehrgpt_args.causal_sfm:
            model_args.max_position_embeddings += 1
        if len(tokenizer.pretrained_token_ids) > 0:
            pretrained_embedding_dim = tokenizer.pretrained_embeddings.shape[1]
        else:
            pretrained_embedding_dim = model_args.hidden_size

        model_args_cehrgpt = model_args.as_dict()
        model_args_cehrgpt.pop("attn_implementation")
        # CEHR-GPT does not support this anymore
        model_args_cehrgpt.pop("exclude_position_ids")
        model_config = CEHRGPTConfig(
            activation_function=cehrgpt_args.activation_function,
            vocab_size=tokenizer.vocab_size,
            value_vocab_size=tokenizer.value_vocab_size,
            time_token_vocab_size=tokenizer.time_token_vocab_size,
            bos_token_id=tokenizer.end_token_id,
            eos_token_id=tokenizer.end_token_id,
            lab_token_ids=tokenizer.lab_token_ids,
            token_to_time_token_mapping=tokenizer.token_to_time_token_mapping,
            attn_implementation=attn_implementation,
            causal_sfm=cehrgpt_args.causal_sfm,
            demographics_size=cehrgpt_args.demographics_size,
            next_token_prediction_loss_weight=cehrgpt_args.next_token_prediction_loss_weight,
            lab_token_penalty=cehrgpt_args.lab_token_penalty,
            lab_token_loss_weight=cehrgpt_args.lab_token_loss_weight,
            value_prediction_loss_weight=cehrgpt_args.value_prediction_loss_weight,
            entropy_penalty=cehrgpt_args.entropy_penalty,
            entropy_penalty_alpha=cehrgpt_args.entropy_penalty_alpha,
            n_pretrained_embeddings_layers=cehrgpt_args.n_pretrained_embeddings_layers,
            use_pretrained_embeddings=len(tokenizer.pretrained_token_ids) > 0,
            pretrained_embedding_dim=pretrained_embedding_dim,
            apply_rotary=cehrgpt_args.apply_rotary,
            sample_packing_max_positions=(
                cehrgpt_args.max_tokens_per_batch
                if cehrgpt_args.sample_packing
                else model_args.max_position_embeddings
            ),
            include_motor_time_to_event=cehrgpt_args.include_motor_time_to_event,
            motor_tte_vocab_size=tokenizer.motor_tte_vocab_size,
            motor_time_to_event_weight=cehrgpt_args.motor_time_to_event_weight,
            motor_num_time_pieces=cehrgpt_args.motor_num_time_pieces,
            ve_token_id=tokenizer.ve_token_id,
            n_inner=cehrgpt_args.inner_dim,
            decoder_mlp=cehrgpt_args.decoder_mlp,
            **model_args_cehrgpt,
        )

    model = CEHRGPT2LMHeadModel(model_config)
    if tokenizer.pretrained_token_ids:
        model.cehrgpt.update_pretrained_embeddings(
            tokenizer.pretrained_token_ids,
            tokenizer.pretrained_embeddings,
        )
    if model.config.torch_dtype == torch.bfloat16:
        return model.bfloat16()
    elif model.config.torch_dtype == torch.float16:
        return model.half()
    return model


def main():
    cehrgpt_args, data_args, model_args, training_args = parse_runner_args()
    if cehrgpt_args.sample_packing and data_args.streaming:
        raise RuntimeError(
            f"sample_packing is not supported when streaming is enabled, please set streaming to False"
        )

    if data_args.streaming:
        # This is for disabling the warning message https://github.com/huggingface/transformers/issues/5486
        # This happens only when streaming is enabled
        os.environ["TOKENIZERS_PARALLELISM"] = "false"
        # The iterable dataset doesn't have sharding implemented, so the number of works has to be set to 0
        # Otherwise the trainer will throw an error
        training_args.dataloader_num_workers = 0
        training_args.dataloader_prefetch_factor = None

    processed_dataset: Optional[DatasetDict] = None
    cache_file_collector = CacheFileCollector()
    if cehrgpt_args.tokenized_dataset_name:
        prepared_ds_path = Path(
            os.path.join(
                data_args.dataset_prepared_path, cehrgpt_args.tokenized_dataset_name
            )
        )
        if prepared_ds_path.exists():
            LOG.warning(
                "The dataset name %s already exists under %s",
                cehrgpt_args.tokenized_dataset_name,
                data_args.dataset_prepared_path,
            )
    else:
        prepared_ds_path = generate_prepared_ds_path(data_args, model_args)

    if os.path.exists(os.path.join(data_args.data_folder, "dataset_dict.json")):
        LOG.info(f"Loading prepared dataset from disk at {data_args.data_folder}...")
        processed_dataset = load_from_disk(data_args.data_folder)
        # If the data has been processed in the past, it's assume the tokenizer has been created before.
        # we load the CEHR-GPT tokenizer from the output folder, otherwise an exception will be raised.
        tokenizer_name_or_path = os.path.expanduser(
            training_args.output_dir
            if cehrgpt_args.expand_tokenizer
            else model_args.tokenizer_name_or_path
        )
        if not tokenizer_exists(tokenizer_name_or_path):
            raise RuntimeError(
                f"The dataset has been tokenized but the corresponding tokenizer: "
                f"{model_args.tokenizer_name_or_path} does not exist"
            )
        cehrgpt_tokenizer = CehrGptTokenizer.from_pretrained(tokenizer_name_or_path)
    elif any(prepared_ds_path.glob("*")):
        LOG.info(f"Loading prepared dataset from disk at {prepared_ds_path}...")
        processed_dataset = load_from_disk(str(prepared_ds_path))
        LOG.info("Prepared dataset loaded from disk...")
        # If the data has been processed in the past, it's assume the tokenizer has been created before.
        # we load the CEHR-GPT tokenizer from the output folder, otherwise an exception will be raised.
        tokenizer_name_or_path = os.path.expanduser(
            training_args.output_dir
            if cehrgpt_args.expand_tokenizer
            else model_args.tokenizer_name_or_path
        )
        if not tokenizer_exists(tokenizer_name_or_path):
            raise RuntimeError(
                f"The dataset has been tokenized but the corresponding tokenizer: "
                f"{model_args.tokenizer_name_or_path} does not exist"
            )
        cehrgpt_tokenizer = CehrGptTokenizer.from_pretrained(tokenizer_name_or_path)
    else:
        # Only run tokenization and data transformation in the main process in torch distributed training
        # otherwise the multiple processes will create tokenizers at the same time
        if is_main_process(training_args.local_rank):
            # If the data is in the MEDS format, we need to convert it to the CEHR-BERT format
            if data_args.is_data_in_meds:
                meds_extension_path = get_meds_extension_path(
                    data_folder=data_args.data_folder,
                    dataset_prepared_path=data_args.dataset_prepared_path,
                )
                try:
                    LOG.info(
                        "Trying to load the MEDS extension from disk at %s...",
                        meds_extension_path,
                    )
                    dataset = load_from_disk(meds_extension_path)
                    if data_args.streaming:
                        if isinstance(dataset, DatasetDict):
                            dataset = {
                                k: v.to_iterable_dataset(
                                    num_shards=training_args.dataloader_num_workers
                                )
                                for k, v in dataset.items()
                            }
                        else:
                            dataset = dataset.to_iterable_dataset(
                                num_shards=training_args.dataloader_num_workers
                            )
                except FileNotFoundError as e:
                    LOG.warning(e)
                    dataset = create_dataset_from_meds_reader(
                        data_args=data_args,
                        dataset_mappings=[
                            MedToCehrGPTDatasetMapping(
                                data_args=data_args,
                                include_inpatient_hour_token=cehrgpt_args.include_inpatient_hour_token,
                            )
                        ],
                        cache_file_collector=cache_file_collector,
                    )
                    if not data_args.streaming:
                        dataset.save_to_disk(str(meds_extension_path))
                        stats = dataset.cleanup_cache_files()
                        LOG.info(
                            "Clean up the cached files for the cehrgpt dataset transformed from the MEDS: %s",
                            stats,
                        )
                        # Clean up the files created from the data generator
                        cache_file_collector.remove_cache_files()
                        dataset = load_from_disk(str(meds_extension_path))
            else:
                # Load the dataset from the parquet files
                dataset = load_parquet_as_dataset(
                    os.path.expanduser(data_args.data_folder),
                    split="train",
                    streaming=data_args.streaming,
                )
                # If streaming is enabled, we need to manually split the data into train/val
                if data_args.streaming and data_args.validation_split_num:
                    dataset = dataset.shuffle(
                        buffer_size=10_000, seed=training_args.seed
                    )
                    train_set = dataset.skip(data_args.validation_split_num)
                    val_set = dataset.take(data_args.validation_split_num)
                    dataset = DatasetDict({"train": train_set, "validation": val_set})
                elif data_args.validation_split_percentage:
                    dataset = dataset.train_test_split(
                        test_size=data_args.validation_split_percentage,
                        seed=training_args.seed,
                    )
                    dataset = DatasetDict(
                        {"train": dataset["train"], "validation": dataset["test"]}
                    )
                else:
                    raise RuntimeError(
                        f"Can not split the data. If streaming is enabled, validation_split_num needs to be "
                        f"defined, otherwise validation_split_percentage needs to be provided. "
                        f"The current values are:\n"
                        f"validation_split_percentage: {data_args.validation_split_percentage}\n"
                        f"validation_split_num: {data_args.validation_split_num}\n"
                        f"streaming: {data_args.streaming}"
                    )

            # Create the CEHR-GPT tokenizer if it's not available in the output folder
            cehrgpt_tokenizer = load_and_create_tokenizer(
                data_args=data_args,
                model_args=model_args,
                cehrgpt_args=cehrgpt_args,
                dataset=dataset,
            )

            # Retrain the tokenizer in case we want to pretrain the model further using different datasets
            if cehrgpt_args.expand_tokenizer:
                new_tokenizer_path = os.path.expanduser(training_args.output_dir)
                try:
                    cehrgpt_tokenizer = CehrGptTokenizer.from_pretrained(
                        new_tokenizer_path
                    )
                except Exception:
                    cehrgpt_tokenizer = CehrGptTokenizer.expand_trained_tokenizer(
                        cehrgpt_tokenizer=cehrgpt_tokenizer,
                        dataset=dataset["train"],
                        data_args=data_args,
                        concept_name_mapping={},
                        pretrained_concept_embedding_model=PretrainedEmbeddings(
                            cehrgpt_args.pretrained_embedding_path
                        ),
                        apply_entropy_filter=cehrgpt_args.apply_entropy_filter,
                        min_prevalence=cehrgpt_args.min_prevalence,
                    )
                    cehrgpt_tokenizer.save_pretrained(
                        os.path.expanduser(training_args.output_dir)
                    )

            # TODO: temp solution, this column is mixed typed and causes an issue when transforming the data
            if not data_args.streaming:
                all_columns = dataset["train"].column_names
                if "visit_concept_ids" in all_columns:
                    dataset = dataset.remove_columns(["visit_concept_ids"])

            # sort the patient features chronologically and tokenize the data
            processed_dataset = create_cehrgpt_pretraining_dataset(
                dataset=dataset,
                cehrgpt_tokenizer=cehrgpt_tokenizer,
                data_args=data_args,
                cache_file_collector=cache_file_collector,
            )
            # only save the data to the disk if it is not streaming
            if not data_args.streaming:
                processed_dataset.save_to_disk(str(prepared_ds_path))
                stats = processed_dataset.cleanup_cache_files()
                LOG.info(
                    "Clean up the cached files for the cehrgpt pretraining dataset: %s",
                    stats,
                )
            cache_file_collector.remove_cache_files()

        # After main-process-only operations, synchronize all processes to ensure consistency
        if dist.is_available() and dist.is_initialized():
            dist.barrier()

        # Loading tokenizer in all processes in torch distributed training
        tokenizer_name_or_path = os.path.expanduser(
            training_args.output_dir
            if cehrgpt_args.expand_tokenizer
            else model_args.tokenizer_name_or_path
        )
        cehrgpt_tokenizer = CehrGptTokenizer.from_pretrained(tokenizer_name_or_path)
        # Load the dataset from disk again to in torch distributed training
        if not data_args.streaming:
            processed_dataset = load_from_disk(str(prepared_ds_path))

    if processed_dataset is None:
        raise RuntimeError("The processed dataset cannot be None")

    def filter_func(examples):
        if cehrgpt_args.drop_long_sequences:
            return [
                model_args.max_position_embeddings >= _ >= data_args.min_num_tokens
                for _ in examples["num_of_concepts"]
            ]
        else:
            return [_ >= data_args.min_num_tokens for _ in examples["num_of_concepts"]]

    # Create the args for batched filtering
    filter_args = {"batched": True, "batch_size": data_args.preprocessing_batch_size}
    # If the dataset is not in a streaming mode, we could add num_proc to enable parallelization
    if not data_args.streaming:
        filter_args["num_proc"] = data_args.preprocessing_num_workers

    # The filter can't be applied to a DatasetDict of IterableDataset (in case of streaming)
    # we need to iterate through all the datasets and apply the filter separately
    if isinstance(processed_dataset, DatasetDict) or isinstance(
        processed_dataset, IterableDatasetDict
    ):
        for key in processed_dataset.keys():
            processed_dataset[key] = processed_dataset[key].filter(
                filter_func, **filter_args
            )
    else:
        processed_dataset = processed_dataset.filter(filter_func, **filter_args)

    model = load_and_create_model(model_args, cehrgpt_args, cehrgpt_tokenizer)

    # Try to update motor tte vocab size if the new configuration is different from the existing one
    if cehrgpt_args.include_motor_time_to_event:
        model.update_motor_tte_vocab_size(cehrgpt_tokenizer.motor_tte_vocab_size)

    # Expand tokenizer to adapt to the new pretraining dataset
    if model.config.vocab_size < cehrgpt_tokenizer.vocab_size:
        model.resize_token_embeddings(cehrgpt_tokenizer.vocab_size)
        # Update the pretrained embedding weights if they are available
        if model.config.use_pretrained_embeddings:
            model.cehrgpt.update_pretrained_embeddings(
                cehrgpt_tokenizer.pretrained_token_ids,
                cehrgpt_tokenizer.pretrained_embeddings,
            )
        elif cehrgpt_tokenizer.pretrained_token_ids:
            model.config.pretrained_embedding_dim = (
                cehrgpt_tokenizer.pretrained_embeddings.shape[1]
            )
            model.config.use_pretrained_embeddings = True
            model.cehrgpt.initialize_pretrained_embeddings()
            model.cehrgpt.update_pretrained_embeddings(
                cehrgpt_tokenizer.pretrained_token_ids,
                cehrgpt_tokenizer.pretrained_embeddings,
            )

    # Detecting last checkpoint.
    last_checkpoint = get_last_hf_checkpoint(training_args)

    # Set seed before initializing model.
    set_seed(training_args.seed)

    if not data_args.streaming and not cehrgpt_args.sample_packing:
        processed_dataset.set_format("pt")

    callbacks = []
    if cehrgpt_args.use_early_stopping:
        callbacks.append(
            CustomEarlyStoppingCallback(
                model_args.early_stopping_patience,
                cehrgpt_args.early_stopping_threshold,
            )
        )

    if cehrgpt_args.sample_packing:
        trainer_class = partial(
            SamplePackingTrainer,
            max_tokens_per_batch=cehrgpt_args.max_tokens_per_batch,
            max_position_embeddings=model_args.max_position_embeddings,
            train_lengths=processed_dataset["train"]["num_of_concepts"],
            validation_lengths=(
                processed_dataset["validation"]
                if "validation" in processed_dataset
                else processed_dataset["test"]
            )["num_of_concepts"],
        )
        training_args.per_device_train_batch_size = 1
        training_args.per_device_eval_batch_size = 1
        data_collator_fn = partial(
            SamplePackingCehrGptDataCollator,
            cehrgpt_args.max_tokens_per_batch,
            model_args.max_position_embeddings,
        )
    else:
        trainer_class = Trainer
        data_collator_fn = CehrGptDataCollator

    trainer = trainer_class(
        model=model,
        data_collator=data_collator_fn(
            tokenizer=cehrgpt_tokenizer,
            max_length=(
                cehrgpt_args.max_tokens_per_batch
                if cehrgpt_args.sample_packing
                else model_args.max_position_embeddings
            ),
            shuffle_records=data_args.shuffle_records,
            include_ttv_prediction=model_args.include_ttv_prediction,
            use_sub_time_tokenization=model_args.use_sub_time_tokenization,
            include_values=model_args.include_values,
            include_motor_time_to_event=cehrgpt_args.include_motor_time_to_event,
            motor_tte_vocab_size=model.config.motor_tte_vocab_size,
            motor_num_time_pieces=cehrgpt_args.motor_num_time_pieces,
            motor_sampling_probability=cehrgpt_args.motor_sampling_probability,
        ),
        train_dataset=processed_dataset["train"],
        eval_dataset=(
            processed_dataset["validation"]
            if "validation" in processed_dataset
            else processed_dataset["test"]
        ),
        args=training_args,
        callbacks=callbacks,
    )

    checkpoint = None
    if training_args.resume_from_checkpoint is not None:
        checkpoint = training_args.resume_from_checkpoint
    elif last_checkpoint is not None:
        checkpoint = last_checkpoint

    train_result = trainer.train(resume_from_checkpoint=checkpoint)
    trainer.save_model()  # Saves the tokenizer too for easy upload
    metrics = train_result.metrics

    trainer.log_metrics("train", metrics)
    trainer.save_metrics("train", metrics)
    trainer.save_state()


if __name__ == "__main__":
    main()



import logging

import numpy as np
from sklearn.preprocessing import OneHotEncoder
from tqdm import tqdm

LOG = logging.getLogger(__name__)

RANDOM_SEE = 42
NUM_OF_GENDERS = 3
NUM_OF_RACES = 21


def get_demographics(concept_ids):
    year_token, age_token, gender, race = concept_ids[0:4]
    try:
        year = int(year_token[5:])
    except ValueError:
        LOG.error(
            f"{year_token[5:]} cannot be converted to an integer, use the default value 1900"
        )
        year = 1900

    try:
        age = int(age_token[4:])
    except ValueError:
        LOG.error(
            f"{age_token[4:]} cannot be converted to an integer, use the default value 1900"
        )
        age = -1

    return year, age, gender, race


def create_race_encoder(
    train_data_sample, evaluation_data_sample, synthetic_data_sample
):
    race_encoder = OneHotEncoder()
    all_unique_races = np.unique(
        np.concatenate(
            [
                train_data_sample.race.unique(),
                evaluation_data_sample.race.unique(),
                synthetic_data_sample.race.unique(),
            ],
            axis=0,
        )
    )
    race_encoder.fit(all_unique_races[:, np.newaxis])
    return race_encoder


def create_gender_encoder(
    train_data_sample, evaluation_data_sample, synthetic_data_sample
):
    gender_encoder = OneHotEncoder()
    all_unique_genders = np.unique(
        np.concatenate(
            [
                train_data_sample.gender.unique(),
                evaluation_data_sample.gender.unique(),
                synthetic_data_sample.gender.unique(),
            ],
            axis=0,
        )
    )
    gender_encoder.fit(all_unique_genders[:, np.newaxis])
    return gender_encoder


def extract_medical_concepts(concept_ids):
    concept_ids = [_ for _ in concept_ids[4:] if str.isnumeric(_)]
    return list(set(concept_ids))


def create_binary_format(concept_ids, concept_tokenizer):
    indices = np.array(concept_tokenizer.encode(concept_ids)).flatten().astype(int)
    embeddings = np.zeros(concept_tokenizer.vocab_size)
    embeddings.put(indices, 1)
    return embeddings


def extract_common_sensitive_concepts(
    dataset, concept_tokenizer, common_attributes, sensitive_attributes
):
    common_embeddings_list = []
    sensitive_embeddings_list = []
    for _, pat_seq in dataset.concept_ids.items():
        concept_ids = extract_medical_concepts(pat_seq)
        common_concept_ids = [_ for _ in concept_ids if _ in common_attributes]
        sensitive_concept_ids = [_ for _ in concept_ids if _ in sensitive_attributes]

        common_embeddings_list.append(
            create_binary_format(common_concept_ids, concept_tokenizer)
        )
        sensitive_embeddings_list.append(
            create_binary_format(sensitive_concept_ids, concept_tokenizer)
        )
    return np.array(common_embeddings_list), np.array(sensitive_embeddings_list)


def transform_concepts(dataset, concept_tokenizer):
    embedding_list = []
    for _, pat_seq in dataset.concept_ids.items():
        embedding_list.append(
            create_binary_format(extract_medical_concepts(pat_seq), concept_tokenizer)
        )
    return np.asarray(embedding_list)


def scale_age(dataset):
    ages = dataset.concept_ids.apply(
        lambda concept_list: get_demographics(concept_list)[1]
    )
    dataset = dataset[ages >= 0]
    ages = ages[ages >= 0]
    max_age = ages.max()
    dataset["scaled_age"] = ages / max_age
    return dataset


def create_demographics(dataset):
    genders = dataset.concept_ids.apply(
        lambda concept_list: get_demographics(concept_list)[2]
    )
    races = dataset.concept_ids.apply(
        lambda concept_list: get_demographics(concept_list)[3]
    )
    dataset["gender"] = genders
    dataset["race"] = races
    return dataset


def create_vector_representations(
    dataset, concept_tokenizer, gender_encoder, race_encoder
):
    concept_vectors = transform_concepts(dataset, concept_tokenizer)
    gender_vectors = gender_encoder.transform(
        dataset.gender.to_numpy()[:, np.newaxis]
    ).todense()
    race_vectors = race_encoder.transform(
        dataset.race.to_numpy()[:, np.newaxis]
    ).todense()
    age_vectors = dataset.scaled_age.to_numpy()[:, np.newaxis]

    pat_vectors = np.concatenate(
        [age_vectors, gender_vectors, race_vectors, concept_vectors], axis=-1
    )

    return np.asarray(pat_vectors)


def create_vector_representations_for_attribute(
    dataset,
    concept_tokenizer,
    gender_encoder,
    race_encoder,
    common_attributes,
    sensitive_attributes,
):
    common_concept_vectors, sensitive_concept_vectors = (
        extract_common_sensitive_concepts(
            dataset, concept_tokenizer, common_attributes, sensitive_attributes
        )
    )
    gender_vectors = gender_encoder.transform(
        dataset.gender.to_numpy()[:, np.newaxis]
    ).todense()
    race_vectors = race_encoder.transform(
        dataset.race.to_numpy()[:, np.newaxis]
    ).todense()
    age_vectors = dataset.scaled_age.to_numpy()[:, np.newaxis]

    common_pat_vectors = np.concatenate(
        [age_vectors, gender_vectors, race_vectors, common_concept_vectors], axis=-1
    )

    return np.asarray(common_pat_vectors), np.asarray(sensitive_concept_vectors)


def batched_pairwise_euclidean_distance_indices(A, B, batch_size, self_exclude=False):
    # Initialize arrays to hold the minimum distances and indices for each point in A
    min_distances = np.full((A.shape[0],), np.inf)
    min_indices = np.full((A.shape[0],), -1, dtype=int)

    # Iterate over A in batches
    for i in tqdm(range(0, A.shape[0], batch_size), total=A.shape[0] // batch_size + 1):
        end_i = i + batch_size
        A_batch = A[i:end_i]

        # Adjust the identity matrix size based on the actual batch size
        actual_batch_size = A_batch.shape[0]

        # Iterate over B in batches
        for j in range(0, B.shape[0], batch_size):
            end_j = j + batch_size
            B_batch = B[j:end_j]

            # Compute distances between the current batches of A and B
            distances = np.sqrt(
                np.sum(
                    (A_batch[:, np.newaxis, :] - B_batch[np.newaxis, :, :]) ** 2, axis=2
                )
            )

            # Apply the identity matrix to exclude self-matches if required
            if self_exclude and i == j:
                identity_matrix = np.eye(actual_batch_size) * 10e8
                distances += identity_matrix

            # Find the minimum distance and corresponding indices for the A batch
            min_batch_indices = np.argmin(distances, axis=1) + j
            min_batch_distances = np.min(distances, axis=1)

            # Update the minimum distances and indices if the current batch distances are smaller
            update_mask = min_batch_distances < min_distances[i:end_i]
            min_distances[i:end_i][update_mask] = min_batch_distances[update_mask]
            min_indices[i:end_i][update_mask] = min_batch_indices[update_mask]

    return min_indices


def find_match(source, target, return_index: bool = False):
    a = np.sum(target**2, axis=1).reshape(target.shape[0], 1) + np.sum(
        source.T**2, axis=0
    )
    b = np.dot(target, source.T) * 2
    distance_matrix = a - b
    return (
        np.argmin(distance_matrix, axis=0)
        if return_index
        else np.min(distance_matrix, axis=0)
    )


def find_match_self(source, target, return_index: bool = False):
    a = np.sum(target**2, axis=1).reshape(target.shape[0], 1) + np.sum(
        source.T**2, axis=0
    )
    b = np.dot(target, source.T) * 2
    distance_matrix = a - b
    n_col = np.shape(distance_matrix)[1]

    if return_index:
        min_indices = np.zeros(n_col, dtype=int)
        for i in range(n_col):
            sorted_indices = np.argsort(distance_matrix[:, i])
            min_indices[i] = sorted_indices[1]  # Get index of second smallest value
        return min_indices
    else:
        min_distance = np.zeros(n_col)
        for i in range(n_col):
            sorted_column = np.sort(distance_matrix[:, i])
            min_distance[i] = sorted_column[1]
        return min_distance



import argparse
import os
import select  # Import select for monitoring stdout and stderr
import subprocess
from pathlib import Path

import yaml


def create_arg_parser():
    parser = argparse.ArgumentParser(
        description="Arguments for benchmarking CEHRGPT on ehrshot cohorts"
    )
    parser.add_argument("--cohort_dir", required=True)
    parser.add_argument("--base_yaml_file", required=True)
    parser.add_argument("--output_folder", required=True)
    return parser.parse_args()


if __name__ == "__main__":
    args = create_arg_parser()

    with open(args.base_yaml_file, "rb") as stream:
        base_config = yaml.safe_load(stream)

    for cohort_name in os.listdir(args.cohort_dir):
        if cohort_name.endswith("/"):
            cohort_name = cohort_name[:-1]
        individual_output = os.path.join(args.output_folder, cohort_name)
        if os.path.exists(individual_output):
            continue
        Path(individual_output).mkdir(parents=True, exist_ok=True)
        base_config["data_folder"] = os.path.join(args.cohort_dir, cohort_name, "train")
        base_config["test_data_folder"] = os.path.join(
            args.cohort_dir, cohort_name, "test"
        )
        base_config["output_dir"] = individual_output

        # Write YAML data to a file
        config_path = os.path.join(individual_output, "config.yaml")
        with open(config_path, "w") as yaml_file:
            yaml.dump(base_config, yaml_file, default_flow_style=False)

        command = [
            "python",
            "-u",
            "-m",
            "cehrgpt.runners.hf_cehrgpt_finetune_runner",
            config_path,
        ]

        # Start the subprocess
        with subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
        ) as process:
            while True:
                # Use select to wait for either stdout or stderr to have data
                reads = [process.stdout.fileno(), process.stderr.fileno()]
                ret = select.select(reads, [], [])

                # Read data from stdout and stderr as it becomes available
                for fd in ret[0]:
                    if fd == process.stdout.fileno():
                        line = process.stdout.readline()
                        if line:
                            print(line, end="")
                    elif fd == process.stderr.fileno():
                        line = process.stderr.readline()
                        if line:
                            print(line, end="")

                # Break loop when process finishes
                if process.poll() is not None:
                    break



from abc import ABC, abstractmethod
from datetime import date, datetime
from typing import Union


def fill_datetime(year: int) -> str:
    """Helper function to create an ISO 8601 datetime string for 01/01 of a given year."""
    return datetime.strptime(f"{year}-01-01", "%Y-%m-%d").isoformat()


def fill_start_datetime(d: Union[date, datetime]) -> str:
    """Helper function to create an ISO 8601 string at 00:00:00 (start of day)."""
    if isinstance(d, datetime):
        return d.isoformat()
    elif isinstance(d, date):
        return datetime.combine(d, datetime.min.time()).isoformat()
    raise RuntimeError(f"{type(d)} is not supported by this function.")


def fill_end_datetime(d: Union[date, datetime]) -> str:
    """Helper function to create an ISO 8601 string at 23:59:59.999999 (end of day)."""
    if isinstance(d, datetime):
        return d.isoformat()
    elif isinstance(d, date):
        return datetime.combine(d, datetime.max.time()).isoformat()
    raise RuntimeError(f"{type(d)} is not supported by this function.")


class OmopEntity(ABC):
    """
    Abstract base class that all OMOP CDM entities inherit from.

    Ensures that each entity can export its data as JSON,
    retrieve its schema, and provide a table name.
    """

    @abstractmethod
    def export_as_json(self):
        pass

    @classmethod
    @abstractmethod
    def get_schema(cls):
        pass

    @abstractmethod
    def get_table_name(self):
        pass


# -----------------------------------------------------------------------------
# PERSON
# https://ohdsi.github.io/CommonDataModel/cdm54.html#person
# -----------------------------------------------------------------------------
class Person(OmopEntity):
    def __init__(self, person_id, gender_concept_id, year_of_birth, race_concept_id):
        self._person_id = person_id
        self._gender_concept_id = gender_concept_id
        self._year_of_birth = year_of_birth
        self._race_concept_id = race_concept_id

    def export_as_json(self):
        return {
            "person_id": self._person_id,
            "gender_concept_id": self._gender_concept_id,
            "year_of_birth": self._year_of_birth,
            "month_of_birth": 1,
            "day_of_birth": 1,
            "birth_datetime": fill_datetime(self._year_of_birth),
            "race_concept_id": self._race_concept_id,
            "ethnicity_concept_id": 0,
            "location_id": 0,
            "provider_id": 0,
            "care_site_id": 0,
            "person_source_value": "",
            "gender_source_value": "",
            "gender_source_concept_id": 0,
            "race_source_value": "",
            "race_source_concept_id": 0,
            "ethnicity_source_value": "",
            "ethnicity_source_concept_id": 0,
        }

    @classmethod
    def get_schema(cls):
        return {
            "person_id": int,
            "gender_concept_id": int,
            "year_of_birth": int,
            "month_of_birth": int,
            "day_of_birth": int,
            "birth_datetime": datetime,
            "race_concept_id": int,
            "ethnicity_concept_id": int,
            "location_id": int,
            "provider_id": int,
            "care_site_id": int,
            "person_source_value": str,
            "gender_source_value": str,
            "gender_source_concept_id": int,
            "race_source_value": str,
            "race_source_concept_id": int,
            "ethnicity_source_value": str,
            "ethnicity_source_concept_id": int,
        }

    def get_table_name(self):
        return "person"


# -----------------------------------------------------------------------------
# VISIT_OCCURRENCE
# https://ohdsi.github.io/CommonDataModel/cdm54.html#visit_occurrence
# -----------------------------------------------------------------------------
class VisitOccurrence(OmopEntity):
    def __init__(
        self,
        visit_occurrence_id: int,
        visit_concept_id: int,
        visit_start_datetime: datetime,
        person: Person,
        discharged_to_concept_id: int = 0,
    ):
        self._visit_occurrence_id = visit_occurrence_id
        self._visit_concept_id = visit_concept_id
        self._visit_start_date = visit_start_datetime.date()
        self._visit_start_datetime = fill_start_datetime(visit_start_datetime)
        self._visit_end_date = visit_start_datetime.date()
        self._visit_end_datetime = fill_end_datetime(visit_start_datetime)
        self._person = person
        self._discharged_to_concept_id = discharged_to_concept_id

    def export_as_json(self):
        return {
            "visit_occurrence_id": self._visit_occurrence_id,
            "visit_concept_id": self._visit_concept_id,
            "person_id": self._person._person_id,
            "visit_start_date": self._visit_start_date,
            "visit_start_datetime": self._visit_start_datetime,
            "visit_end_date": self._visit_end_date,
            "visit_end_datetime": self._visit_end_datetime,
            "visit_type_concept_id": 44818702,  # default concept (e.g. Inpatient Visit)
            "provider_id": 0,
            "care_site_id": 0,
            "visit_source_value": "",
            "visit_source_concept_id": self._visit_concept_id,
            "admitted_from_concept_id": 0,  # replaced "admitting_source_concept_id"
            "admitted_from_source_value": "",  # replaced "admitting_source_value"
            "discharged_to_concept_id": self._discharged_to_concept_id,
            "discharged_to_source_value": "",
            "preceding_visit_occurrence_id": 0,
        }

    @classmethod
    def get_schema(cls):
        return {
            "visit_occurrence_id": int,
            "visit_concept_id": int,
            "person_id": int,
            "visit_start_date": date,
            "visit_start_datetime": datetime,
            "visit_end_date": date,
            "visit_end_datetime": datetime,
            "visit_type_concept_id": int,
            "provider_id": int,
            "care_site_id": int,
            "visit_source_value": str,
            "visit_source_concept_id": int,
            "admitted_from_concept_id": int,
            "admitted_from_source_value": str,
            "discharged_to_concept_id": int,
            "discharged_to_source_value": str,
            "preceding_visit_occurrence_id": int,
        }

    def get_table_name(self):
        return "visit_occurrence"

    @property
    def person(self):
        return self._person

    @property
    def discharged_to_concept_id(self):
        return self._discharged_to_concept_id

    def set_discharged_to_concept_id(self, discharged_to_concept_id: int):
        self._discharged_to_concept_id = discharged_to_concept_id

    def set_visit_end_date(self, visit_end_datetime: datetime):
        self._visit_end_date = visit_end_datetime.date()
        self._visit_end_datetime = fill_end_datetime(visit_end_datetime)


# -----------------------------------------------------------------------------
# CONDITION_OCCURRENCE
# https://ohdsi.github.io/CommonDataModel/cdm54.html#condition_occurrence
# -----------------------------------------------------------------------------
class ConditionOccurrence(OmopEntity):
    def __init__(
        self,
        condition_occurrence_id: int,
        condition_concept_id: int,
        visit_occurrence: VisitOccurrence,
        condition_datetime: datetime,
    ):
        self._condition_occurrence_id = condition_occurrence_id
        self._condition_concept_id = condition_concept_id
        self._visit_occurrence = visit_occurrence
        self._condition_start_date = condition_datetime.date()
        self._condition_start_datetime = fill_start_datetime(condition_datetime)
        self._condition_end_date = condition_datetime.date()
        self._condition_end_datetime = fill_start_datetime(condition_datetime)

    def export_as_json(self):
        return {
            "condition_occurrence_id": self._condition_occurrence_id,
            "person_id": self._visit_occurrence._person._person_id,
            "condition_concept_id": self._condition_concept_id,
            "condition_start_date": self._condition_start_date,
            "condition_start_datetime": self._condition_start_datetime,
            "condition_end_date": self._condition_end_date,
            "condition_end_datetime": self._condition_end_datetime,
            "condition_type_concept_id": 32817,  # default concept
            "stop_reason": "",
            "provider_id": 0,
            "visit_occurrence_id": self._visit_occurrence._visit_occurrence_id,
            "visit_detail_id": 0,
            "condition_source_value": "",
            "condition_source_concept_id": self._condition_concept_id,
            "condition_status_source_value": "",
            "condition_status_concept_id": 0,
        }

    @classmethod
    def get_schema(cls):
        return {
            "condition_occurrence_id": int,
            "person_id": int,
            "condition_concept_id": int,
            "condition_start_date": date,
            "condition_start_datetime": datetime,
            "condition_end_date": date,
            "condition_end_datetime": datetime,
            "condition_type_concept_id": int,
            "stop_reason": str,
            "provider_id": int,
            "visit_occurrence_id": int,
            "visit_detail_id": int,
            "condition_source_value": str,
            "condition_source_concept_id": int,
            "condition_status_source_value": str,
            "condition_status_concept_id": int,
        }

    def get_table_name(self):
        return "condition_occurrence"


# -----------------------------------------------------------------------------
# DRUG_EXPOSURE
# https://ohdsi.github.io/CommonDataModel/cdm54.html#drug_exposure
# -----------------------------------------------------------------------------
class DrugExposure(OmopEntity):
    def __init__(
        self,
        drug_exposure_id: int,
        drug_concept_id: int,
        visit_occurrence: VisitOccurrence,
        drug_datetime: datetime,
    ):
        self._drug_exposure_id = drug_exposure_id
        self._drug_concept_id = drug_concept_id
        self._visit_occurrence = visit_occurrence
        self._drug_exposure_start_date = drug_datetime.date()
        self._drug_exposure_start_datetime = fill_start_datetime(drug_datetime)
        self._drug_exposure_end_date = drug_datetime.date()
        self._drug_exposure_end_datetime = fill_start_datetime(drug_datetime)

    def export_as_json(self):
        return {
            "drug_exposure_id": self._drug_exposure_id,
            "person_id": self._visit_occurrence._person._person_id,
            "drug_concept_id": self._drug_concept_id,
            "drug_exposure_start_date": self._drug_exposure_start_date,
            "drug_exposure_start_datetime": self._drug_exposure_start_datetime,
            "drug_exposure_end_date": self._drug_exposure_end_date,
            "drug_exposure_end_datetime": self._drug_exposure_end_datetime,
            "verbatim_end_date": self._drug_exposure_end_date,
            "drug_type_concept_id": 38000177,  # default concept
            "stop_reason": "",
            "refills": None,
            "quantity": None,
            "days_supply": None,
            "sig": "",
            "route_concept_id": 0,
            "lot_number": "",
            "provider_id": 0,
            "visit_occurrence_id": self._visit_occurrence._visit_occurrence_id,
            "visit_detail_id": 0,
            "drug_source_value": "",
            "drug_source_concept_id": self._drug_concept_id,
            "route_source_value": "",
            "dose_unit_source_value": "",
        }

    @classmethod
    def get_schema(cls):
        return {
            "drug_exposure_id": int,
            "person_id": int,
            "drug_concept_id": int,
            "drug_exposure_start_date": date,
            "drug_exposure_start_datetime": datetime,
            "drug_exposure_end_date": date,
            "drug_exposure_end_datetime": datetime,
            "verbatim_end_date": date,
            "drug_type_concept_id": int,
            "stop_reason": str,
            "refills": int,
            "quantity": int,
            "days_supply": int,
            "sig": str,
            "route_concept_id": int,
            "lot_number": str,
            "provider_id": int,
            "visit_occurrence_id": int,
            "visit_detail_id": int,
            "drug_source_value": str,
            "drug_source_concept_id": int,
            "route_source_value": str,
            "dose_unit_source_value": str,
        }

    def get_table_name(self):
        return "drug_exposure"


# -----------------------------------------------------------------------------
# PROCEDURE_OCCURRENCE
# https://ohdsi.github.io/CommonDataModel/cdm54.html#procedure_occurrence
# -----------------------------------------------------------------------------
class ProcedureOccurrence(OmopEntity):
    def __init__(
        self,
        procedure_occurrence_id: int,
        procedure_concept_id: int,
        visit_occurrence: VisitOccurrence,
        procedure_datetime: datetime,
    ):
        self._procedure_occurrence_id = procedure_occurrence_id
        self._procedure_concept_id = procedure_concept_id
        self._visit_occurrence = visit_occurrence
        self._procedure_date = procedure_datetime.date()
        self._procedure_datetime = fill_start_datetime(procedure_datetime)

    def export_as_json(self):
        return {
            "procedure_occurrence_id": self._procedure_occurrence_id,
            "person_id": self._visit_occurrence._person._person_id,
            "procedure_concept_id": self._procedure_concept_id,
            "procedure_date": self._procedure_date,
            "procedure_datetime": self._procedure_datetime,
            "procedure_type_concept_id": 38000178,  # default
            "modifier_concept_id": 0,
            "quantity": 1,
            "provider_id": 0,
            "visit_occurrence_id": self._visit_occurrence._visit_occurrence_id,
            "visit_detail_id": 0,
            "procedure_source_value": "",
            "procedure_source_concept_id": self._procedure_concept_id,
            "qualifier_source_value": "",
        }

    @classmethod
    def get_schema(cls):
        return {
            "procedure_occurrence_id": int,
            "person_id": int,
            "procedure_concept_id": int,
            "procedure_date": date,
            "procedure_datetime": datetime,
            "procedure_type_concept_id": int,
            "modifier_concept_id": int,
            "quantity": int,
            "provider_id": int,
            "visit_occurrence_id": int,
            "visit_detail_id": int,
            "procedure_source_value": str,
            "procedure_source_concept_id": int,
            "qualifier_source_value": str,
        }

    def get_table_name(self):
        return "procedure_occurrence"


# -----------------------------------------------------------------------------
# DEATH
# https://ohdsi.github.io/CommonDataModel/cdm54.html#death
# -----------------------------------------------------------------------------
class Death(OmopEntity):
    def __init__(
        self,
        person: Person,
        death_date: date,
        death_type_concept_id: int = 0,
    ):
        self._person = person
        self._death_date = death_date
        self._death_datetime = fill_end_datetime(death_date)
        self._death_type_concept_id = death_type_concept_id

    def export_as_json(self):
        return {
            "person_id": self._person._person_id,
            "death_date": self._death_date,
            "death_datetime": self._death_datetime,
            "death_type_concept_id": self._death_type_concept_id,
            "cause_concept_id": 0,
            "cause_source_value": "",
            "cause_source_concept_id": 0,
        }

    @classmethod
    def get_schema(cls):
        return {
            "person_id": int,
            "death_date": date,
            "death_datetime": datetime,
            "death_type_concept_id": int,
            "cause_concept_id": int,
            "cause_source_value": str,
            "cause_source_concept_id": int,
        }

    def get_table_name(self):
        return "death"


# -----------------------------------------------------------------------------
# MEASUREMENT
# https://ohdsi.github.io/CommonDataModel/cdm54.html#measurement
# -----------------------------------------------------------------------------
class Measurement(OmopEntity):
    def __init__(
        self,
        measurement_id: int,
        measurement_concept_id: int,
        value_as_number: float,
        is_numeric_type,
        value_as_concept_id,
        visit_occurrence: VisitOccurrence,
        measurement_datetime: datetime,
        unit_source_value,
    ):
        self._measurement_id = measurement_id
        self._measurement_concept_id = measurement_concept_id
        self._value_as_number = value_as_number
        self._value_as_concept_id = value_as_concept_id
        self._visit_occurrence = visit_occurrence
        self._measurement_date = measurement_datetime.date()
        self._measurement_datetime = fill_start_datetime(measurement_datetime)
        self._operator_concept_id = 4172703 if is_numeric_type == 1 else 0
        self._unit_source_value = unit_source_value

    def export_as_json(self):
        return {
            "measurement_id": self._measurement_id,
            "person_id": self._visit_occurrence._person._person_id,
            "measurement_concept_id": self._measurement_concept_id,
            "measurement_date": self._measurement_date,
            "measurement_datetime": self._measurement_datetime,
            "measurement_type_concept_id": 0,
            "operator_concept_id": self._operator_concept_id,
            "value_as_number": self._value_as_number,
            "value_as_concept_id": self._value_as_concept_id,
            "unit_concept_id": 0,
            "range_low": None,
            "range_high": None,
            "provider_id": 0,
            "visit_occurrence_id": self._visit_occurrence._visit_occurrence_id,
            "visit_detail_id": 0,
            "measurement_source_value": "",
            "measurement_source_concept_id": self._measurement_concept_id,
            "unit_source_value": self._unit_source_value,
            "value_source_value": "",
        }

    @classmethod
    def get_schema(cls):
        return {
            "measurement_id": int,
            "person_id": int,
            "measurement_concept_id": int,
            "measurement_date": date,
            "measurement_datetime": datetime,
            "measurement_type_concept_id": int,
            "operator_concept_id": int,
            "value_as_number": float,
            "value_as_concept_id": int,
            "unit_concept_id": int,
            "range_low": float,
            "range_high": float,
            "provider_id": int,
            "visit_occurrence_id": int,
            "visit_detail_id": int,
            "measurement_source_value": str,
            "measurement_source_concept_id": int,
            "unit_source_value": str,
            "value_source_value": str,
        }

    def get_table_name(self):
        return "measurement"



import logging
import os
import random
import sys
from datetime import datetime
from typing import Union

import numpy as np
import pandas as pd
import yaml

from cehrgpt.models.tokenization_hf_cehrgpt import CehrGptTokenizer

from .utils import (
    batched_pairwise_euclidean_distance_indices,
    create_demographics,
    create_gender_encoder,
    create_race_encoder,
    create_vector_representations_for_attribute,
    find_match,
    find_match_self,
    scale_age,
)

RANDOM_SEE = 42
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)
LOG = logging.getLogger("attribute_inference")


def safe_divide(numerator_vector, denominator_vector):
    return np.where(denominator_vector > 0, numerator_vector / denominator_vector, 0)


def cal_f1_score(vector_a, vector_b, index_matrix):
    # vector_a is train data and vector_b is synthetic data or iteself
    shared_vector = np.logical_and(
        vector_a[: len(index_matrix)], vector_b[index_matrix]
    ).astype(int)
    shared_vector_sum = np.sum(shared_vector, axis=1)

    precision = safe_divide(shared_vector_sum, np.sum(vector_b, axis=1))
    recall = safe_divide(shared_vector_sum, np.sum(vector_a, axis=1))

    f1 = safe_divide(2 * recall * precision, recall + precision)
    return f1, precision, recall


def main(args):
    try:
        with open(args.attribute_config, "r") as file:
            data = yaml.safe_load(file)
        if "common_attributes" in data:
            common_attributes = data["common_attributes"]
        if "sensitive_attributes" in data:
            sensitive_attributes = data["sensitive_attributes"]
    except Union[FileNotFoundError, PermissionError, OSError] as e:
        sys.exit(e)

    attribute_inference_folder = os.path.join(args.output_folder, "attribute_inference")
    if not os.path.exists(attribute_inference_folder):
        LOG.info(
            f"Creating the attribute_inference output folder at {attribute_inference_folder}"
        )
        os.makedirs(attribute_inference_folder, exist_ok=True)

    LOG.info(f"Started loading tokenizer at {args.tokenizer_path}")
    concept_tokenizer = CehrGptTokenizer.from_pretrained(args.tokenizer_path)

    LOG.info(f"Started loading training data at {args.training_data_folder}")
    train_data = pd.read_parquet(args.training_data_folder)

    LOG.info(f"Started loading synthetic_data at {args.synthetic_data_folder}")
    synthetic_data = pd.read_parquet(args.synthetic_data_folder)

    LOG.info(
        "Started extracting the demographic information from the patient sequences"
    )
    train_data = create_demographics(train_data)
    synthetic_data = create_demographics(synthetic_data)

    LOG.info("Started rescaling age columns")
    train_data = scale_age(train_data)
    synthetic_data = scale_age(synthetic_data)

    LOG.info("Started encoding gender")
    gender_encoder = create_gender_encoder(
        train_data,
        # TODO need to change this function to be generic
        train_data[:10],
        synthetic_data,
    )
    LOG.info("Completed encoding gender")

    LOG.info("Started encoding race")
    race_encoder = create_race_encoder(
        train_data,
        # TODO need to change this function to be generic
        train_data[:10],
        synthetic_data,
    )
    LOG.info("Completed encoding race")

    random.seed(RANDOM_SEE)
    for i in range(1, args.n_iterations + 1):
        LOG.info(f"Iteration {i}: Started creating data samples")
        train_data_sample = train_data.sample(args.num_of_samples)
        synthetic_data_sample = synthetic_data.sample(args.num_of_samples)
        LOG.info(f"Iteration {i}: Started creating train sample vectors")
        train_common_vectors, train_sensitive_vectors = (
            create_vector_representations_for_attribute(
                train_data_sample,
                concept_tokenizer,
                gender_encoder,
                race_encoder,
                common_attributes=common_attributes,
                sensitive_attributes=sensitive_attributes,
            )
        )

        LOG.info(f"Iteration {i}: Started creating synthetic vectors")
        synthetic_common_vectors, synthetic_sensitive_vectors = (
            create_vector_representations_for_attribute(
                synthetic_data_sample,
                concept_tokenizer,
                gender_encoder,
                race_encoder,
                common_attributes=common_attributes,
                sensitive_attributes=sensitive_attributes,
            )
        )

        LOG.info(
            f"Started calculating the distances between synthetic and training vectors"
        )
        if args.batched:
            train_synthetic_index = batched_pairwise_euclidean_distance_indices(
                train_common_vectors,
                synthetic_common_vectors,
                batch_size=args.batch_size,
            )
            train_train_index = batched_pairwise_euclidean_distance_indices(
                train_common_vectors,
                train_common_vectors,
                batch_size=args.batch_size,
                self_exclude=True,
            )
        else:
            train_synthetic_index = find_match(
                train_common_vectors, synthetic_common_vectors, return_index=True
            )
            train_train_index = find_match_self(
                train_common_vectors, train_common_vectors, return_index=True
            )

        f1_syn_train, precision_syn_train, recall_syn_train = cal_f1_score(
            train_sensitive_vectors, synthetic_sensitive_vectors, train_synthetic_index
        )
        f1_train_train, precision_train_train, recall_train_train = cal_f1_score(
            train_sensitive_vectors, train_sensitive_vectors, train_train_index
        )

        results = {
            "Precision Synthetic Train": precision_syn_train,
            "Recall Synthetic Train": recall_syn_train,
            "F1 Synthetic Train": f1_syn_train,
            "Precision Train Train": precision_train_train,
            "Recall Train Train": recall_train_train,
            "F1 Train Train": f1_train_train,
        }
        LOG.info(
            f"Attribute Inference: Average Precision Synthetic Train: {np.mean(precision_syn_train)} \n"
            f"Attribute Inference: Average Recall Synthetic Train:{np.mean(recall_syn_train)} \n"
            f"Attribute Inference: Average F1 Synthetic Train: {np.mean(f1_syn_train)} \n"
            f"Attribute Inference: Average Precision Train Train: {np.mean(precision_train_train)} \n"
            f"Attribute Inference: Average Recall Train Train: {np.mean(recall_train_train)} \n"
            f"Attribute Inference: Average F1 Train Train: {np.mean(f1_train_train)}"
        )
        current_time = datetime.now().strftime("%m-%d-%Y-%H-%M-%S")
        pd.DataFrame(
            results,
            columns=[
                "Precision Synthetic Train",
                "Recall Synthetic Train",
                "F1 Synthetic Train",
                "Precision Train Train",
                "Recall Train Train",
                "F1 Train Train",
            ],
        ).to_parquet(
            os.path.join(
                attribute_inference_folder,
                f"attribute_inference_{current_time}.parquet",
            )
        )


def create_argparser():
    import argparse

    parser = argparse.ArgumentParser(
        description="Attribute Inference Analysis Arguments"
    )
    parser.add_argument(
        "--training_data_folder",
        dest="training_data_folder",
        action="store",
        help="The path for where the training data folder",
        required=True,
    )
    parser.add_argument(
        "--synthetic_data_folder",
        dest="synthetic_data_folder",
        action="store",
        help="The path for where the synthetic data folder",
        required=True,
    )
    parser.add_argument(
        "--output_folder",
        dest="output_folder",
        action="store",
        help="The output folder that stores the metrics",
        required=True,
    )
    parser.add_argument(
        "--tokenizer_path",
        dest="tokenizer_path",
        action="store",
        help="The path to ConceptTokenizer",
        required=True,
    )
    parser.add_argument(
        "--attribute_config",
        dest="attribute_config",
        action="store",
        help="The configuration yaml file for common and sensitive attributes",
        required=True,
    )
    parser.add_argument(
        "--batch_size",
        dest="batch_size",
        action="store",
        type=int,
        default=1000,
        help="The batch size of the matching algorithm",
        required=False,
    )
    parser.add_argument(
        "--batched",
        dest="batched",
        action="store_true",
        help="Indicate whether we want to use the batch matrix operation",
    )
    parser.add_argument(
        "--num_of_samples",
        dest="num_of_samples",
        action="store",
        type=int,
        required=False,
        default=5000,
    )
    parser.add_argument(
        "--n_iterations",
        dest="n_iterations",
        action="store",
        type=int,
        required=False,
        default=1,
    )
    return parser


if __name__ == "__main__":
    main(create_argparser().parse_args())



import random
from typing import Any, Dict, List, Optional

import numpy as np
import pandas as pd
import torch
from cehrbert.data_generators.hf_data_generator.hf_dataset_mapping import DatasetMapping
from transformers.utils import logging

from cehrgpt.gpt_utils import (
    DEMOGRAPHIC_PROMPT_SIZE,
    collect_demographic_prompts_at_visits,
    construct_age_sequence,
    construct_time_sequence,
    extract_time_interval_in_days,
    extract_time_interval_in_hours,
    is_att_token,
    is_clinical_event,
    is_inpatient_att_token,
    is_inpatient_hour_token,
    random_slice_gpt_sequence,
)
from cehrgpt.models.tokenization_hf_cehrgpt import CehrGptTokenizer

TIME_TO_EVENT_MAX_TIME = 3650
INPATIENT_STAY_DURATION_LIMIT = 30
LOG = logging.get_logger("transformers")


class CehrGptDataProcessor(DatasetMapping):
    def __init__(
        self,
        tokenizer: CehrGptTokenizer,
        max_length: int,
        shuffle_records: bool = False,
        include_values: bool = False,
        include_ttv_prediction: bool = False,
        include_motor_time_to_event: bool = False,
        motor_sampling_probability: float = 0.5,
        pretraining: bool = True,
        include_demographics: bool = False,
        add_linear_prob_token: bool = False,
    ):
        self.tokenizer = tokenizer
        self.max_length = max_length

        self.vs_token_id = tokenizer.vs_token_id
        self.ve_token_id = tokenizer.ve_token_id

        self.shuffle_records = shuffle_records
        self.include_values = include_values
        self.include_ttv_prediction = include_ttv_prediction
        self.pretraining = pretraining
        self.include_demographics = include_demographics
        self.add_linear_prob_token = add_linear_prob_token
        self.empty_array = np.asarray([])

        if self.pretraining and self.add_linear_prob_token:
            raise ValueError(
                "pretraining and add_linear_prob_token cannot be specify at the same time"
            )

        # Motor related codes
        self.include_motor_time_to_event = include_motor_time_to_event
        self.motor_sampling_probability = motor_sampling_probability
        self.motor_code_cache: Dict[str, List[str]] = {}
        # Pre-compute vocab-wide token type mappings
        self._precompute_vocab_mappings()

    def _precompute_vocab_mappings(self):
        """Pre-compute token type mappings for entire vocabulary."""
        LOG.info("Pre-computing vocabulary-wide token mappings...")

        vocab = self.tokenizer.get_vocab()
        self.vocab_to_idx = {token: idx for idx, token in enumerate(vocab.keys())}
        self.vocab_tokens = list(vocab.keys())

        # Pre-compute boolean arrays for token types
        n_vocab = len(self.vocab_tokens)
        self.is_att_token_array = np.zeros(n_vocab, dtype=bool)
        self.is_clinical_event_array = np.zeros(n_vocab, dtype=bool)
        self.time_intervals_array = np.full(n_vocab, -1, dtype=int)

        for i, token in enumerate(self.vocab_tokens):
            if is_att_token(token):
                self.is_att_token_array[i] = True
                try:
                    self.time_intervals_array[i] = extract_time_interval_in_days(token)
                except (ValueError, AttributeError):
                    self.time_intervals_array[i] = -1

            if is_clinical_event(token):
                self.is_clinical_event_array[i] = True

        LOG.info(f"Processed {n_vocab} vocabulary tokens")

    @staticmethod
    def _convert_time_to_event(concept_ids):
        def default_value(c):
            try:
                if is_att_token(c):
                    time_to_visit = extract_time_interval_in_days(c)
                    if (
                        is_inpatient_att_token(c)
                        and time_to_visit > INPATIENT_STAY_DURATION_LIMIT
                    ):
                        return -100
                    return time_to_visit
                elif is_inpatient_hour_token(c):
                    return extract_time_interval_in_hours(c) / 24
                return -100
            except ValueError:
                return -100

        return [float(default_value(_)) for _ in concept_ids]

    def random_sort(self, record: Dict[str, Any]) -> Dict[str, Any]:
        if "record_ranks" not in record:
            return record

        sorting_column = record["record_ranks"]
        random_order = np.random.rand(len(sorting_column))

        if self.include_values:
            iterator = zip(
                sorting_column,
                random_order,
                record["input_ids"],
                record["value_indicators"],
                record["values"],
            )
            sorted_list = sorted(iterator, key=lambda tup2: (tup2[0], tup2[1], tup2[2]))
            _, _, sorted_input_ids, sorted_value_indicators, sorted_values = zip(
                *list(sorted_list)
            )
            record["input_ids"] = sorted_input_ids
            record["value_indicators"] = sorted_value_indicators
            record["values"] = sorted_values
        else:
            iterator = zip(sorting_column, random_order, record["input_ids"])
            sorted_list = sorted(iterator, key=lambda tup2: (tup2[0], tup2[1], tup2[2]))
            _, _, sorted_input_ids = zip(*list(sorted_list))
            record["input_ids"] = sorted_input_ids
        return record

    def transform(self, example: Dict[str, Any]) -> Dict[str, Any]:

        if self.shuffle_records:
            example = self.random_sort(example)

        if "concept_ids" not in example:
            input_ids = example["input_ids"]
            if isinstance(input_ids, torch.Tensor):
                input_ids = input_ids.detach().tolist()
            example["concept_ids"] = self.tokenizer.decode(
                input_ids, skip_special_tokens=False
            )
        example["ages"] = pd.Series(example["ages"]).ffill().tolist()
        example = self.slice_out_input_sequence(example)
        # Add the motor labels
        if self.include_motor_time_to_event:
            motor_inputs = self.create_time_to_event_labels(example)
            example.update(motor_inputs)
        del example["concept_ids"]
        return example

    def update_inputs_based_on_indexes(
        self,
        record: Dict[str, Any],
        start_index,
        end_index,
        add_end_token: bool = False,
        demographic_tokens: Optional[List[str]] = None,
    ) -> Dict[str, Any]:

        last_token_id = (
            self.tokenizer.linear_token_id
            if self.add_linear_prob_token
            else self.tokenizer.end_token_id
        )

        add_last_token = self.add_linear_prob_token | add_end_token

        # Slice out the concept ids
        record["concept_ids"] = (
            (demographic_tokens if demographic_tokens is not None else [])
            + (record["concept_ids"][start_index:end_index])
            + (
                self.tokenizer.decode([last_token_id], skip_special_tokens=False)
                if add_last_token
                else []
            )
        )

        record["input_ids"] = np.concatenate(
            [
                (
                    np.asarray(self.tokenizer.encode(demographic_tokens))
                    if demographic_tokens is not None
                    else self.empty_array
                ),
                np.asarray(record["input_ids"][start_index:end_index]),
                (np.asarray([last_token_id]) if add_last_token else self.empty_array),
            ]
        ).astype(np.int32)

        record["ages"] = np.concatenate(
            [
                (
                    np.full(
                        [DEMOGRAPHIC_PROMPT_SIZE],
                        record["ages"][0] if len(record["ages"]) > 0 else 0,
                    )
                    if demographic_tokens is not None
                    else self.empty_array
                ),
                np.asarray(record["ages"][start_index:end_index]),
                (
                    (
                        np.asarray([record["ages"][-1]])
                        if len(record["ages"]) > 0
                        else self.empty_array
                    )
                    if add_last_token
                    else self.empty_array
                ),
            ]
        ).astype(np.int32)

        # For the new datasets, they contain the column "epoch_times"
        record["epoch_times"] = np.concatenate(
            [
                (
                    np.full(
                        [DEMOGRAPHIC_PROMPT_SIZE],
                        (
                            record["epoch_times"][0]
                            if len(record["epoch_times"]) > 0
                            else 0
                        ),
                    )
                    if demographic_tokens is not None
                    else self.empty_array
                ),
                np.asarray(record["epoch_times"][start_index:end_index]),
                (
                    (
                        np.asarray([record["epoch_times"][-1]])
                        if len(record["epoch_times"]) > 0
                        else self.empty_array
                    )
                    if add_last_token
                    else self.empty_array
                ),
            ]
        ).astype(np.float32)

        if self.include_values:
            record["value_indicators"] = np.concatenate(
                [
                    (
                        np.zeros([DEMOGRAPHIC_PROMPT_SIZE])
                        if demographic_tokens is not None
                        else self.empty_array
                    ),
                    np.asarray(record["value_indicators"][start_index:end_index]),
                    np.asarray([False]) if add_last_token else self.empty_array,
                ]
            ).astype(np.bool_)
            record["values"] = np.concatenate(
                [
                    (
                        np.full(
                            [DEMOGRAPHIC_PROMPT_SIZE], self.tokenizer.pad_value_token_id
                        )
                        if demographic_tokens is not None
                        else self.empty_array
                    ),
                    np.asarray(record["values"][start_index:end_index]),
                    (
                        np.asarray([self.tokenizer.pad_value_token_id])
                        if add_last_token
                        else self.empty_array
                    ),
                ]
            ).astype(np.int32)

        if self.include_ttv_prediction:
            record["time_to_visits"] = np.concatenate(
                [
                    (
                        np.full([DEMOGRAPHIC_PROMPT_SIZE], -100.0)
                        if demographic_tokens is not None
                        else self.empty_array
                    ),
                    np.asarray(record["time_to_visits"][start_index:end_index]),
                    np.asarray([-100.0]) if add_last_token else self.empty_array,
                ]
            ).astype(np.float32)

        return record

    def slice_out_input_sequence(self, record: Dict[str, Any]) -> Dict[str, Any]:
        """Adding the start and end indices to extract a portion of the patient sequence."""
        # Subtract one for the [END] or [LINEAR_PROB] token when sample_packing is not enabled
        new_max_length = (
            self.max_length - 1
            if self.add_linear_prob_token or self.pretraining
            else self.max_length
        )
        concept_ids = record["concept_ids"]
        seq_length = len(record["input_ids"])

        # For backward compatibility, in case these two columns do not already exist
        record["ages"] = construct_age_sequence(record["concept_ids"], record["ages"])
        record["epoch_times"] = construct_time_sequence(
            record["concept_ids"], record["epoch_times"]
        )

        if self.include_ttv_prediction:
            record["time_to_visits"] = np.asarray(
                self._convert_time_to_event(record["concept_ids"])
            )

        # Return the record directly if the actual sequence length is less than the max sequence
        if seq_length <= new_max_length:
            # We only add [END] to the end of the sequence in pre-training
            record = self.update_inputs_based_on_indexes(
                record, 0, seq_length, add_end_token=self.pretraining
            )
            return record

        if self.pretraining:
            # There is a 50% chance we randomly slice out a portion of the patient history and update the demographic
            # prompt depending on the new starting point
            if random.random() < 0.5:
                start_index, end_index, demographic_tokens = random_slice_gpt_sequence(
                    concept_ids, new_max_length
                )
                if start_index != end_index:
                    record = self.update_inputs_based_on_indexes(
                        record, start_index, end_index + 1, add_end_token=False
                    )
                    return record

            end_index = new_max_length - 1
            # The default employs a right truncation strategy, where the demographic prompt is reserved
            for i in reversed(list(range(0, end_index))):
                current_token = record["input_ids"][i]
                if current_token == self.ve_token_id:
                    # Plus one because slicing is right exclusive
                    end_index = i + 1
                    break

            record = self.update_inputs_based_on_indexes(
                record=record, start_index=0, end_index=end_index, add_end_token=False
            )
            return record
        else:
            if self.include_demographics:
                # We employ a left truncation strategy, where the most recent patient history is reserved for fine-tuning
                demographic_prompts_at_visits = collect_demographic_prompts_at_visits(
                    concept_ids
                )
                for token_index, demographic_prompt in demographic_prompts_at_visits:
                    if (
                        seq_length - token_index
                        <= new_max_length - DEMOGRAPHIC_PROMPT_SIZE
                    ):
                        return self.update_inputs_based_on_indexes(
                            record=record,
                            start_index=token_index,
                            end_index=seq_length,
                            add_end_token=False,
                            demographic_tokens=demographic_prompt,
                        )
            else:
                start_index = seq_length - new_max_length
                end_index = seq_length
                for i in range(start_index, end_index):
                    current_token = record["input_ids"][i]
                    if current_token == self.vs_token_id:
                        return self.update_inputs_based_on_indexes(
                            record=record,
                            start_index=i,
                            end_index=end_index,
                            add_end_token=False,
                        )

            # This could happen when the last visit contains more than new_max_length number of tokens
            # We simply take the last new_max_length number of tokens from the patient sequence
            if len(record["input_ids"]) > new_max_length:
                record = self.update_inputs_based_on_indexes(
                    record=record,
                    start_index=-new_max_length,
                    end_index=seq_length,
                    add_end_token=False,
                )
            return record

    def create_time_to_event_labels(self, record: Dict[str, Any]) -> Dict[str, Any]:
        """
        Generates time-to-event (TTE) labels and censoring indicators for each visit in a patient's timeline.

        Processes the input sequence in reverse to compute the number of days from each visit (marked by [VE])
        to the occurrence of future motor-related events.

        Args:
            record (Dict[str, Any]): A dictionary containing the encoded patient sequence with the key "input_ids".
                This sequence includes [VS], [VE], time delta tokens, and motor TTE concept codes.

        Returns:
            Dict[str, Any]: The updated input record with added keys:
                - "time_to_event_vectors": np.ndarray of shape [num_visits, motor_vocab_size], containing time-to-event values
                - "event_indicators": np.ndarray of shape [num_visits, motor_vocab_size], where 0 = event occurred, 1 = censored
        """

        """Highly optimized vectorized version using pre-computed token type arrays."""
        concept_ids = record["concept_ids"]
        # Convert concept_ids to indices for vectorized operations
        concept_indices = np.array([self.vocab_to_idx[cid] for cid in concept_ids])
        # Vectorized token type detection
        is_att_tokens = self.is_att_token_array[concept_indices]
        is_clinical_events = self.is_clinical_event_array[concept_indices]
        time_intervals = self.time_intervals_array[concept_indices]

        # Find valid time tokens (att tokens with positive intervals)
        valid_time_tokens = is_att_tokens & (time_intervals > 0)
        n_concepts = len(concept_ids)

        # We need to make sure event_times is monotonic
        event_times = np.zeros(n_concepts, dtype=float)
        previous_time_stamp = record["epoch_times"][0]
        for i, time_stamp in enumerate(record["epoch_times"]):
            if time_stamp < previous_time_stamp:
                time_stamp = previous_time_stamp
            else:
                previous_time_stamp = time_stamp
            event_times[i] = time_stamp

        # Determine prediction positions
        before_valid_time_tokens = np.roll(valid_time_tokens, -1)
        # We randomly make predictions at 50% of the sequence positions
        prediction_positions = (
            np.random.rand(n_concepts) < self.motor_sampling_probability
        )
        # We don't predict at the att time tokens
        prediction_positions &= ~is_att_tokens
        # We disable TTE predictions using the demographics alone
        prediction_positions[:4] = False
        # We take the union of the random prediction positions and the positions right before time token
        prediction_positions = prediction_positions | before_valid_time_tokens
        # We exclude the events that occur at the last time stamp
        prediction_positions &= event_times != event_times[-1]

        prediction_indices = np.where(prediction_positions)[0]
        if len(prediction_indices) == 0:
            return {
                "motor_censor_times": [],
                "motor_row_indices": [],
                "motor_col_indices": [],
                "motor_values": [],
                "motor_tte_task_indicators": [False] * n_concepts,
            }

        # Pre-compute all motor codes for clinical events to avoid repeated lookups
        clinical_positions = np.where(is_clinical_events)[0]
        motor_codes_cache = {}  # position -> list of (motor_code, motor_token_id)

        for pos in clinical_positions:
            concept_id = concept_ids[pos]
            if concept_id in self.motor_code_cache:
                motor_codes = self.motor_code_cache[concept_id]
            else:
                motor_codes = self.tokenizer.get_motor_parents(concept_id)
                self.motor_code_cache[concept_id] = motor_codes

            if motor_codes:
                motor_codes_cache[pos] = [
                    (motor_code, self.tokenizer.get_motor_token_id(motor_code))
                    for motor_code in motor_codes
                ]

        # Process sections in REVERSE order but build results in FORWARD order
        section_boundaries = np.concatenate([prediction_indices, [n_concepts]])
        last_event_time = event_times[-1]

        # Pre-allocate arrays with exact size needed
        num_prediction_positions = len(prediction_indices)
        motor_censor_times = np.zeros(num_prediction_positions, dtype=float)
        motor_tte_task_indicators = np.zeros(n_concepts, dtype=bool)

        # Store sparse matrix data grouped by row for efficient construction
        sparse_data_by_row = {}  # row_idx -> [(col_idx, value), ...]

        # Global motor event state that accumulates as we go backwards
        global_motor_events = {}  # motor_code -> earliest_future_time

        # Process in reverse order but assign to forward row indices
        for i in range(len(prediction_indices) - 1, -1, -1):
            start_index = prediction_indices[i]
            end_index = section_boundaries[i + 1]
            current_event_time = event_times[start_index]

            # Add new motor events from this section to global state
            section_start = start_index + 1
            section_end = end_index + 1 if end_index < n_concepts else n_concepts

            # Process clinical events in this section (in reverse order within section)
            section_clinical_positions = clinical_positions[
                (clinical_positions >= section_start)
                & (clinical_positions < section_end)
            ]

            for pos in reversed(section_clinical_positions):
                if pos in motor_codes_cache:
                    concept_time = event_times[pos]
                    if concept_time > current_event_time:
                        for motor_code, motor_token_id in motor_codes_cache[pos]:
                            global_motor_events[motor_code] = (
                                concept_time,
                                motor_token_id,
                            )

            # Store sparse matrix data for current prediction position
            # Even if global_motor_events is empty, we still need to record this position
            # because it indicates all motor tasks are censored at this time point
            sparse_data_by_row[i] = [
                (motor_token_id, motor_time - current_event_time)
                for motor_code, (
                    motor_time,
                    motor_token_id,
                ) in global_motor_events.items()
            ]
            motor_tte_task_indicators[start_index] = True
            motor_censor_times[i] = last_event_time - current_event_time

        # Build final sparse matrix lists in forward order (no reversals needed)
        motor_row_indices = []
        motor_col_indices = []
        motor_values = []

        for row_idx in sorted(sparse_data_by_row.keys()):
            for col_idx, value in sparse_data_by_row[row_idx]:
                motor_row_indices.append(row_idx)
                motor_col_indices.append(col_idx)
                motor_values.append(value)

        # Filter out unused positions from motor_censor_times
        motor_censor_times = [
            motor_censor_times[i] for i in sorted(sparse_data_by_row.keys())
        ]

        if len(motor_row_indices) == 0:
            LOG.debug(
                "No MOTOR tasks detected for this sample. "
                "Length: %s, last 10 concepts: %s",
                len(concept_ids),
                concept_ids[-10:] if len(concept_ids) >= 10 else concept_ids,
            )

        return {
            "motor_censor_times": motor_censor_times,
            "motor_row_indices": motor_row_indices,
            "motor_col_indices": motor_col_indices,
            "motor_values": motor_values,
            "motor_tte_task_indicators": motor_tte_task_indicators.tolist(),
        }



from typing import Iterator, List, Optional

import numpy as np
import torch
import torch.distributed as dist
from torch.utils.data import Sampler
from transformers import logging

LOG = logging.get_logger("transformers")


class SamplePlacerHolder:
    def __init__(self):
        self.epoch = 0

    def set_epoch(self, epoch):
        self.epoch = epoch


class SamplePackingBatchSampler(Sampler[List[int]]):
    """
    A batch sampler that creates batches by packing samples together.

    to maximize GPU utilization, ensuring the total tokens per batch
    doesn't exceed max_tokens.
    """

    def __init__(
        self,
        lengths: List[int],
        max_tokens_per_batch: int,
        max_position_embeddings: int,
        num_replicas: Optional[int] = None,
        rank: Optional[int] = None,
        seed: int = 0,
        drop_last: bool = False,
        negative_sampling_probability: Optional[float] = None,
        labels: Optional[List[int]] = None,
    ):
        """
        Args:

            lengths: List of sequence lengths for each sample
            max_tokens: Maximum number of tokens in a batch
            drop_last: Whether to drop the last incomplete batch
        """
        super().__init__()

        if num_replicas is None:
            if dist.is_available() and dist.is_initialized():
                num_replicas = dist.get_world_size()
                LOG.info(
                    "torch.distributed is initialized and there are %s of replicas",
                    num_replicas,
                )
            else:
                num_replicas = 1
                LOG.info(
                    "torch.dist is not initialized and therefore default to 1 for num_replicas"
                )

        if rank is None:
            if dist.is_available() and dist.is_initialized():
                rank = dist.get_rank()
                LOG.info(
                    "torch.distributed is initialized and the current rank is %s", rank
                )
            else:
                rank = 0
                LOG.info(
                    "torch.distributed is not initialized and therefore default to 0 for rank"
                )

        if not (0 <= rank < num_replicas):
            raise ValueError(
                f"Invalid rank {rank}, rank should be in the interval [0, {num_replicas - 1}]"
            )

        if negative_sampling_probability is not None and labels is None:
            raise ValueError(
                f"When the negative sampling probability is provide, the labels must be provided as well"
            )

        self.lengths = lengths
        self.max_tokens_per_batch = max_tokens_per_batch
        self.max_position_embeddings = max_position_embeddings
        self.num_replicas = num_replicas
        self.rank = rank
        self.seed = seed
        self.drop_last = drop_last
        self.negative_sampling_probability = negative_sampling_probability
        self.labels = labels
        # Trainer https://github.com/huggingface/transformers/blame/main/src/transformers/trainer.py#L2470
        # http://github.com/huggingface/accelerate/blob/v0.31.0/src/accelerate/data_loader.py#L482
        # the huggingface trainer will call the accelerate.data_loader.DataLoaderShard.set_epoch,
        # which will call batch_sampler.sample.set_epoch
        self.sampler = SamplePlacerHolder()

    def __iter__(self) -> Iterator[List[int]]:

        # deterministically shuffle based on epoch and seed
        g = torch.Generator()
        g.manual_seed(self.seed + self.sampler.epoch)
        indices = torch.randperm(len(self.lengths), generator=g).tolist()

        # Partition indices for this rank
        indices = indices[self.rank :: self.num_replicas]

        batch = []
        current_batch_tokens = 0

        for idx in indices:
            # There is a chance to skip the negative samples to account for the class imbalance
            # in the fine-tuning dataset
            if self.negative_sampling_probability:
                if (
                    np.random.random() > self.negative_sampling_probability
                    and self.labels[idx] == 0
                ):
                    continue
            # We take the minimum of the two because each sequence will be truncated to fit
            # the context window of the model
            sample_length = min(self.lengths[idx], self.max_position_embeddings)
            # If adding this sample would exceed max_tokens_per_batch, yield the current batch
            if (
                current_batch_tokens + sample_length + 2 > self.max_tokens_per_batch
                and batch
            ):
                yield batch
                batch = []
                current_batch_tokens = 0

            # Add the sample to the current batch
            batch.append(idx)
            # plus extract one for the [END] and [PAD] tokens to separate samples
            current_batch_tokens += sample_length + 2

        # Yield the last batch if it's not empty and we're not dropping it
        if batch and not self.drop_last:
            yield batch

    def __len__(self) -> int:
        """
        Estimates the number of batches that will be generated.

        This is an approximation since the exact number depends on the specific
        sequence lengths and their order.
        """
        if len(self.lengths) == 0:
            return 0

        # There is a chance to skip the negative samples to account for the class imbalance
        # in the fine-tuning dataset
        if self.negative_sampling_probability:
            truncated_lengths = []
            for length, label in zip(self.lengths, self.labels):
                if (
                    np.random.random() > self.negative_sampling_probability
                    and label == 0
                ):
                    continue
                truncated_lengths.append(length)
        else:
            # We need to truncate the lengths due to the context window limit imposed by the model
            truncated_lengths = [
                min(self.max_position_embeddings, length + 2) for length in self.lengths
            ]

        # Calculate average sequence length
        avg_seq_length = sum(truncated_lengths) // len(truncated_lengths)

        # Estimate average number of sequences per batch
        seqs_per_batch = self.max_tokens_per_batch // avg_seq_length

        # Estimate total number of batches
        if self.drop_last:
            # If dropping last incomplete batch
            return len(truncated_lengths) // seqs_per_batch
        else:
            # If keeping last incomplete batch, ensure at least 1 batch
            return max(1, len(truncated_lengths) // seqs_per_batch)



from typing import Optional, Tuple

import numpy as np
import torch
import torch.optim as optim
from sklearn.metrics import accuracy_score, roc_auc_score
from torch.nn import CrossEntropyLoss
from transformers import BertConfig, BertModel

from cehrgpt.simulations.time_embedding_simulation import generate_simulation_data


class ModelTimeToken(torch.nn.Module):
    def __init__(self, vocab_size: int):
        super(ModelTimeToken, self).__init__()
        self.embedding = torch.nn.Embedding(vocab_size, 16)
        self.bert = BertModel(
            BertConfig(
                vocab_size=vocab_size,
                hidden_size=16,
                num_attention_heads=2,
                num_hidden_layers=2,
                intermediate_size=32,
                hidden_dropout_prob=0.0,
                attention_probs_dropout_prob=0.0,
                max_position_embeddings=3,
            ),
            add_pooling_layer=False,
        )
        self.linear = torch.nn.Linear(48, 2)

    def forward(
        self,
        input_ids: torch.LongTensor,
        labels: Optional[torch.LongTensor] = None,
    ) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
        bz = input_ids.shape[0]
        x = self.embedding(input_ids)
        bert_output = self.bert.forward(inputs_embeds=x, return_dict=True)
        output = bert_output.last_hidden_state.reshape((bz, 48))
        y = self.linear(output)
        loss = None
        if labels is not None:
            loss_fct = CrossEntropyLoss()
            loss = loss_fct(y, labels)
        return loss, y


def create_time_token_tokenizer(simulated_data):
    vocab = []
    for row in simulated_data:
        x1, x2, t1, t2, y = row
        x1 = f"c-{x1}"
        x2 = f"c-{x2}"
        t = f"t-{t2 - t1}"
        if x1 not in vocab:
            vocab.append(x1)
        if x2 not in vocab:
            vocab.append(x2)
        if t not in vocab:
            vocab.append(t)
    return {c: i + 1 for i, c in enumerate(vocab)}


device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


def eval_step(simulated_data, time_token_tokenizer, time_embedding_model):
    time_embedding_model.eval()
    eval_input_ids = []
    eval_y = []
    for row in simulated_data:
        x1, x2, t1, t2, y = row
        x1 = f"c-{x1}"
        x2 = f"c-{x2}"
        t = f"t-{t2 - t1}"
        eval_input_ids.append(
            [
                time_token_tokenizer[x1],
                time_token_tokenizer[t],
                time_token_tokenizer[x2],
            ]
        )
        eval_y.append(y)
    with torch.no_grad():
        batched_input_ids = torch.tensor(eval_input_ids, dtype=torch.long).to(device)
        batched_y = np.asarray(eval_y)
        # Compute loss and forward pass
        _, y_pred = time_embedding_model(batched_input_ids)
        y_probs = torch.nn.functional.softmax(y_pred, dim=1)
        y_probs = y_probs.detach().cpu().numpy()
        roc_auc = roc_auc_score(batched_y, y_probs[:, 1])
        accuracy = accuracy_score(batched_y, y_probs[:, 1] > y_probs[:, 0])
        print(f"ROC AUC: {roc_auc}")
        print(f"Accuracy: {accuracy}")
    return accuracy, roc_auc


def train_step(
    simulated_data, time_token_tokenizer, time_embedding_model, time_embedding_optimizer
):
    batched_input_ids = []
    batched_y = []
    indices = np.random.choice(simulated_data.shape[0], size=8, replace=False)
    for row in simulated_data[indices, :]:
        x1, x2, t1, t2, y = row
        x1 = f"c-{x1}"
        x2 = f"c-{x2}"
        t = f"t-{t2 - t1}"
        batched_input_ids.append(
            [
                time_token_tokenizer[x1],
                time_token_tokenizer[t],
                time_token_tokenizer[x2],
            ]
        )
        batched_y.append(y)
    batched_input_ids = torch.tensor(batched_input_ids, dtype=torch.long).to(device)
    batched_y = torch.tensor(batched_y, dtype=torch.long).to(device)
    # Zero the gradients
    time_embedding_optimizer.zero_grad()
    # Compute loss and forward pass
    loss, _ = time_embedding_model(batched_input_ids, batched_y)
    # Backward pass (compute gradients)
    loss.backward()
    # Update model parameters
    time_embedding_optimizer.step()
    return loss


def main(args):
    simulated_data = generate_simulation_data(args.n_samples)
    time_token_tokenizer = create_time_token_tokenizer(simulated_data)
    time_embedding_model = ModelTimeToken(len(time_token_tokenizer) + 1).to(device)
    time_embedding_optimizer = optim.Adam(time_embedding_model.parameters(), lr=0.001)
    steps = []
    roc_aucs = []
    accuracies = []
    for step in range(args.n_steps):
        loss = train_step(
            simulated_data,
            time_token_tokenizer,
            time_embedding_model,
            time_embedding_optimizer,
        )
        print(f"Step {step}: Loss = {loss.item()}")
        # Evaluation
        if (
            args.n_steps % args.eval_frequency == 0
            and args.n_steps > args.eval_frequency
        ):
            accuracy, roc_auc = eval_step(
                simulated_data, time_token_tokenizer, time_embedding_model
            )
            steps.append(step)
            roc_aucs.append(roc_auc)
            accuracies.append(accuracy)
    return {"steps": steps, "roc_auc": roc_aucs, "accuracy": accuracies}


if __name__ == "__main__":
    import argparse
    import json
    from pathlib import Path

    parser = argparse.ArgumentParser("Model with time token simulation")
    parser.add_argument("--output_dir", type=str, required=True)
    parser.add_argument("--n_steps", type=int, default=10000)
    parser.add_argument("--n_samples", type=int, default=1000)
    parser.add_argument("--batch_size", type=int, default=128)
    parser.add_argument("--eval_frequency", type=int, default=100)
    args = parser.parse_args()
    output_dir = Path(args.output_dir)
    output_dir.mkdir(exist_ok=True, parents=True)
    metrics = main(args)
    with open(output_dir / "time_token_metrics.json", "w") as f:
        json.dump(metrics, f)



from typing import Optional, Tuple, Union

import torch.nn.functional as f
import torch.utils.checkpoint
from torch import nn
from transformers.activations import ACT2FN
from transformers.models.gpt2.modeling_gpt2 import Conv1D, GPT2Attention
from transformers.utils import is_flash_attn_2_available, logging

if is_flash_attn_2_available():
    from flash_attn import flash_attn_func, flash_attn_varlen_func
    from flash_attn.bert_padding import index_first_axis, pad_input, unpad_input  # noqa

from cehrgpt.models.activations import RMSNorm

logger = logging.get_logger("transformers")


def is_sample_pack(attention_mask: torch.Tensor) -> bool:
    """
    Determines whether any sequence in the batch is likely sample-packed.

    A sample-packed sequence is one where there are non-padding (1) tokens
    after a padding (0) token, indicating multiple sequences packed together
    with padding as a separator.

    Args:
        attention_mask (torch.Tensor): A tensor of shape (batch_size, seq_len)
            where 1 indicates a real token and 0 indicates padding.

    Returns:
        bool: True if any sample in the batch is sample-packed, False otherwise.
    """

    # If the attention_maks is left padded, we will flip it so we can use the same logic below
    if (attention_mask[:, 0] == 0).any():
        attention_mask = attention_mask.flip(dims=[1])

    nonzero_counts = attention_mask.sum(dim=1)
    max_token_positions = torch.argmax(
        attention_mask.to(torch.int32).flip(dims=[1]), dim=1
    )
    max_indices = attention_mask.shape[1] - 1 - max_token_positions
    return torch.any(nonzero_counts < (max_indices + 1)).item()


# Copied from transformers.models.llama.modeling_llama._get_unpad_data
def _get_unpad_data(attention_mask):
    # This infers sample packing
    if is_sample_pack(attention_mask):
        # Assume input: attention_mask shape = (batch, seq_len)
        attention_mask = attention_mask.flatten()  # shape: (seq_len,)

        # Compute max_index of the last non-zero element
        nonzero = torch.nonzero(attention_mask, as_tuple=False).flatten()
        max_index = nonzero[-1].item()

        # Pad the truncated attention mask
        padded_attention_mask = f.pad(attention_mask[: max_index + 1], (0, 1), value=0)

        # Indices of all tokens
        indices = torch.nonzero(attention_mask, as_tuple=False).flatten()

        # Find where 0s occur (segment boundaries)
        cumsum_seqlens_in_batch = torch.cumsum(padded_attention_mask, dim=0)[
            padded_attention_mask == 0
        ]

        # Compute seqlens per segment
        seqlens_in_batch = (
            cumsum_seqlens_in_batch
            - f.pad(cumsum_seqlens_in_batch, (1, 0), value=0)[:-1]
        ).to(torch.int)

        max_seqlen_in_batch = (
            seqlens_in_batch.max().item() if seqlens_in_batch.numel() > 0 else 0
        )
        cu_seqlens = f.pad(cumsum_seqlens_in_batch, (1, 0)).to(torch.int)
    else:
        seqlens_in_batch = attention_mask.sum(dim=-1, dtype=torch.int32)
        indices = torch.nonzero(attention_mask.flatten(), as_tuple=False).flatten()
        max_seqlen_in_batch = seqlens_in_batch.max().item()
        cu_seqlens = f.pad(
            torch.cumsum(seqlens_in_batch, dim=0, dtype=torch.int32), (1, 0)
        )

    return (
        indices,
        cu_seqlens,
        max_seqlen_in_batch,
    )


class RotaryPositionEmbedding(nn.Module):
    def __init__(self, dim: int):
        super().__init__()
        self.dim = dim
        self.inv_freq = 1.0 / (10000 ** (torch.linspace(0, 2, steps=dim // 2))).reshape(
            1, 1, dim // 2
        )

    def forward(self, x: torch.Tensor, time: torch.Tensor) -> torch.Tensor:
        if time.ndim == 2:
            time = time[..., None]
        t = self.inv_freq.to(time.device) * time
        sin, cos = torch.sin(t), torch.cos(t)
        sin = torch.stack((sin, sin), dim=-1).reshape(x.shape)
        cos = torch.stack((cos, cos), dim=-1).reshape(x.shape)
        flat_x = x.reshape(-1, x.shape[-1])
        x1 = flat_x[:, ::2]
        x2 = flat_x[:, 1::2]
        return (x * cos) + (torch.stack((-x2, x1), dim=-1).reshape(x.shape) * sin)


class GPT2AttentionRoPE(GPT2Attention):
    """
    GPT2FlashAttention inherits from `GPT2Attention`.

    The primary change is in the forward pass, where it correctly
    calls the public API of flash attention and handles padding tokens.
    """

    def __init__(
        self, config, is_cross_attention=False, layer_idx=None, apply_rotary=False
    ):
        super().__init__(config, is_cross_attention, layer_idx)
        self.apply_rotary = apply_rotary
        if self.apply_rotary:
            self.rope = RotaryPositionEmbedding(config.hidden_size)

    def forward(
        self,
        hidden_states: Optional[Tuple[torch.FloatTensor]],
        position_ids: Optional[Tuple[torch.FloatTensor]] = None,
        layer_past: Optional[Tuple[torch.Tensor]] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.FloatTensor] = None,
        use_cache: Optional[bool] = False,
        output_attentions: Optional[bool] = False,
    ) -> Tuple[Union[torch.Tensor, Tuple[torch.Tensor]], ...]:

        if encoder_hidden_states is not None:
            if not hasattr(self, "q_attn"):
                raise ValueError(
                    "If class is used as cross attention, the weights `q_attn` have to be defined. "
                    "Please make sure to instantiate class with `GPT2Attention(..., is_cross_attention=True)`."
                )

            query = self.q_attn(hidden_states)
            key, value = self.c_attn(encoder_hidden_states).split(
                self.split_size, dim=2
            )
            attention_mask = encoder_attention_mask
        else:
            query, key, value = self.c_attn(hidden_states).split(self.split_size, dim=2)

        if self.apply_rotary and position_ids is not None:
            query = self.rope(query, position_ids)
            key = self.rope(key, position_ids)
            # value = self.rope(value, position_ids)

        query = self._split_heads(query, self.num_heads, self.head_dim)
        key = self._split_heads(key, self.num_heads, self.head_dim)
        value = self._split_heads(value, self.num_heads, self.head_dim)

        if layer_past is not None:
            past_key, past_value = layer_past
            key = torch.cat((past_key, key), dim=-2)
            value = torch.cat((past_value, value), dim=-2)

        if use_cache is True:
            present = (key, value)
        else:
            present = None

        if self.reorder_and_upcast_attn:
            attn_output, attn_weights = self._upcast_and_reordered_attn(
                query, key, value, attention_mask, head_mask
            )
        else:
            attn_output, attn_weights = self._attn(
                query, key, value, attention_mask, head_mask
            )

        attn_output = self._merge_heads(attn_output, self.num_heads, self.head_dim)
        attn_output = self.c_proj(attn_output)
        attn_output = self.resid_dropout(attn_output)

        outputs = (attn_output, present)
        if output_attentions:
            outputs += (attn_weights,)

        return outputs  # a, present, (attentions)


class GPT2FlashAttention(GPT2Attention):
    """
    GPT2FlashAttention inherits from `GPT2Attention`.

    The primary change is in the forward pass, where it correctly
    calls the public API of flash attention and handles padding tokens.
    """

    def __init__(
        self, config, is_cross_attention=False, layer_idx=None, apply_rotary=False
    ):
        super().__init__(config, is_cross_attention, layer_idx)
        self.apply_rotary = apply_rotary
        if self.apply_rotary:
            self.rope = RotaryPositionEmbedding(config.hidden_size)

    def forward(
        self,
        hidden_states: Optional[Tuple[torch.FloatTensor]],
        position_ids: Optional[Tuple[torch.FloatTensor]] = None,
        layer_past: Optional[Tuple[torch.Tensor]] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.FloatTensor] = None,
        use_cache: Optional[bool] = False,
        output_attentions: Optional[bool] = False,
    ) -> Tuple[Union[torch.Tensor, Tuple[torch.Tensor]], ...]:

        # Prepare query, key, and value
        if encoder_hidden_states is not None:
            if not hasattr(self, "q_attn"):
                raise ValueError(
                    "If class is used as cross attention, the weights `q_attn` have to be defined. "
                    "Please make sure to instantiate class with `GPT2Attention(..., is_cross_attention=True)`."
                )

            query = self.q_attn(hidden_states)
            key, value = self.c_attn(encoder_hidden_states).split(
                self.split_size, dim=2
            )
            attention_mask = encoder_attention_mask
        else:
            query, key, value = self.c_attn(hidden_states).split(self.split_size, dim=2)

        if self.apply_rotary and position_ids is not None:
            query = self.rope(query, position_ids)
            key = self.rope(key, position_ids)
            # value = self.rope(value, position_ids)

        query = self._split_heads(query, self.num_heads, self.head_dim)
        key = self._split_heads(key, self.num_heads, self.head_dim)
        value = self._split_heads(value, self.num_heads, self.head_dim)

        if layer_past is not None:
            past_key, past_value = layer_past
            key = torch.cat((past_key, key), dim=-2)
            value = torch.cat((past_value, value), dim=-2)

        if use_cache is True:
            present = (key, value)
        else:
            present = None

        # Apply Flash Attention Forward
        if self.reorder_and_upcast_attn:
            attn_output, attn_weights = self._upcast_and_reordered_attn(
                query, key, value, attention_mask, head_mask
            )
        else:
            # Flash Attention forward pass
            attn_output = self._flash_attention_forward(
                query,
                key,
                value,
                attention_mask,
                query.size(-2),
                self.attn_dropout.p,
                softmax_scale=None,
            )

        # Merge heads and project back to hidden size
        attn_output = self._merge_heads(attn_output, self.num_heads, self.head_dim)
        attn_output = self.c_proj(attn_output)
        attn_output = self.resid_dropout(attn_output)

        outputs = (attn_output, present)
        if output_attentions:
            outputs += (attn_weights,)

        return outputs

    def _flash_attention_forward(
        self,
        query_states,
        key_states,
        value_states,
        attention_mask,
        query_length,
        dropout=0.0,
        softmax_scale=None,
    ):
        """
        Calls the forward method of Flash Attention - if the input hidden states contain at least one padding token.

        first unpad the input, then computes the attention scores and pad the final attention scores.
        Args:
            query_states (`torch.Tensor`):
                Input query states to be passed to Flash Attention API
            key_states (`torch.Tensor`):
                Input key states to be passed to Flash Attention API
            value_states (`torch.Tensor`):
                Input value states to be passed to Flash Attention API
            attention_mask (`torch.Tensor`):
                The padding mask - corresponds to a tensor of size `(batch_size, seq_len)` where 0 stands for the
                position of padding tokens and 1 for the position of non-padding tokens.
            dropout (`int`, *optional*):
                Attention dropout
            softmax_scale (`float`, *optional*):
                The scaling of QK^T before applying softmax. Default to 1 / sqrt(head_dim)
        """

        # Flash attention requires the input to have the shape
        # batch_size x seq_length x head_dim x hidden_dim
        # therefore we just need to keep the original shape
        dtype = query_states.dtype
        query_states = query_states.permute(0, 2, 1, 3).contiguous().to(torch.bfloat16)
        key_states = key_states.permute(0, 2, 1, 3).contiguous().to(torch.bfloat16)
        value_states = value_states.permute(0, 2, 1, 3).contiguous().to(torch.bfloat16)

        # Contains at least one padding token in the sequence
        if attention_mask is not None:
            batch_size = query_states.shape[0]

            (
                query_states,
                key_states,
                value_states,
                indices_q,
                cu_seq_lens,
                max_seq_lens,
            ) = self._upad_input(
                query_states, key_states, value_states, attention_mask, query_length
            )

            cu_seqlens_q, cu_seqlens_k = cu_seq_lens
            max_seqlen_in_batch_q, max_seqlen_in_batch_k = max_seq_lens

            attn_output_unpad = flash_attn_varlen_func(
                query_states,
                key_states,
                value_states,
                cu_seqlens_q=cu_seqlens_q,
                cu_seqlens_k=cu_seqlens_k,
                max_seqlen_q=max_seqlen_in_batch_q,
                max_seqlen_k=max_seqlen_in_batch_k,
                dropout_p=dropout,
                softmax_scale=softmax_scale,
                causal=True,
            )
            # (batch, seq_length, n_heads, head_dim)
            attn_output = pad_input(
                attn_output_unpad, indices_q, batch_size, query_length
            )
        else:
            attn_output = flash_attn_func(
                query_states,
                key_states,
                value_states,
                dropout,
                softmax_scale=softmax_scale,
                causal=self.is_causal,
            )
        # re-order the tensor back to (batch, n_heads, seq_length, head_dim)
        return attn_output.permute(0, 2, 1, 3).contiguous().to(dtype)

    # Copied from transformers.models.llama.modeling_llama.LlamaFlashAttention2._upad_input
    def _upad_input(
        self, query_layer, key_layer, value_layer, attention_mask, query_length
    ):
        indices_k, cu_seqlens_k, max_seqlen_in_batch_k = _get_unpad_data(attention_mask)
        batch_size, kv_seq_len, num_key_value_heads, head_dim = key_layer.shape

        key_layer = index_first_axis(
            key_layer.reshape(batch_size * kv_seq_len, num_key_value_heads, head_dim),
            indices_k,
        )
        value_layer = index_first_axis(
            value_layer.reshape(batch_size * kv_seq_len, num_key_value_heads, head_dim),
            indices_k,
        )
        if query_length == kv_seq_len:
            query_layer = index_first_axis(
                query_layer.reshape(batch_size * kv_seq_len, self.num_heads, head_dim),
                indices_k,
            )
            cu_seqlens_q = cu_seqlens_k
            max_seqlen_in_batch_q = max_seqlen_in_batch_k
            indices_q = indices_k
        elif query_length == 1:
            max_seqlen_in_batch_q = 1
            cu_seqlens_q = torch.arange(
                batch_size + 1, dtype=torch.int32, device=query_layer.device
            )  # There is a memcpy here, that is very bad.
            indices_q = cu_seqlens_q[:-1]
            query_layer = query_layer.squeeze(1)
        else:
            # The -q_len: slice assumes left padding.
            attention_mask = attention_mask[:, -query_length:]
            query_layer, indices_q, cu_seqlens_q, max_seqlen_in_batch_q = unpad_input(
                query_layer, attention_mask
            )

        return (
            query_layer,
            key_layer,
            value_layer,
            indices_q,
            (cu_seqlens_q, cu_seqlens_k),
            (max_seqlen_in_batch_q, max_seqlen_in_batch_k),
        )


class LlamaMLP(nn.Module):
    def __init__(self, intermediate_size, config):
        super().__init__()
        self.config = config
        self.hidden_size = config.hidden_size
        self.intermediate_size = intermediate_size
        self.gate_proj = nn.Linear(
            self.hidden_size, self.intermediate_size, bias=config.mlp_bias
        )
        self.up_proj = nn.Linear(
            self.hidden_size, self.intermediate_size, bias=config.mlp_bias
        )
        self.down_proj = nn.Linear(
            self.intermediate_size, self.hidden_size, bias=config.mlp_bias
        )
        self.act_fn = ACT2FN[config.activation_function]

    def forward(self, x: Optional[Tuple[torch.FloatTensor]]) -> torch.FloatTensor:
        down_proj = self.down_proj(self.act_fn(self.gate_proj(x)) * self.up_proj(x))
        return down_proj


class GPT2MLP(nn.Module):
    def __init__(self, intermediate_size, config):
        super().__init__()
        embed_dim = config.hidden_size
        self.c_fc = Conv1D(intermediate_size, embed_dim)
        self.c_proj = Conv1D(embed_dim, intermediate_size)
        self.act = ACT2FN[config.activation_function]
        self.dropout = nn.Dropout(config.resid_pdrop)

    def forward(
        self, hidden_states: Optional[Tuple[torch.FloatTensor]]
    ) -> torch.FloatTensor:
        hidden_states = self.c_fc(hidden_states)
        hidden_states = self.act(hidden_states)
        hidden_states = self.c_proj(hidden_states)
        hidden_states = self.dropout(hidden_states)
        return hidden_states


class GPT2Block(nn.Module):
    def __init__(self, config, layer_idx=None):
        super().__init__()
        hidden_size = config.hidden_size
        inner_dim = config.n_inner if config.n_inner is not None else 4 * hidden_size
        attention_class = (
            GPT2FlashAttention
            if getattr(config, "_attn_implementation", "eager") == "flash_attention_2"
            else GPT2AttentionRoPE
        )
        self.ln_1 = nn.LayerNorm(hidden_size, eps=config.layer_norm_epsilon)
        self.attn = attention_class(
            config=config, layer_idx=layer_idx, apply_rotary=config.apply_rotary
        )
        self.ln_2 = nn.LayerNorm(hidden_size, eps=config.layer_norm_epsilon)

        if config.add_cross_attention:
            self.crossattention = attention_class(
                config=config, is_cross_attention=True, layer_idx=layer_idx
            )
            self.ln_cross_attn = nn.LayerNorm(
                hidden_size, eps=config.layer_norm_epsilon
            )

        decoder_mlp_function = getattr(config, "decoder_mlp", "GPT2MLP")
        if decoder_mlp_function == "GPT2MLP":
            self.mlp = GPT2MLP(inner_dim, config)
        elif getattr(config, "decoder_mlp", "GPT2Block") == "LlamaMLP":
            self.mlp = LlamaMLP(inner_dim, config)
        else:
            raise RuntimeError("You must set decoder_mlp to one of (GPT2MLP, LlamaMLP)")

    def forward(
        self,
        hidden_states: Optional[Tuple[torch.FloatTensor]],
        position_ids: Optional[Tuple[torch.FloatTensor]] = None,
        layer_past: Optional[Tuple[torch.Tensor]] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.FloatTensor] = None,
        use_cache: Optional[bool] = False,
        output_attentions: Optional[bool] = False,
    ) -> Union[
        Tuple[torch.Tensor],
        Optional[Tuple[torch.Tensor, Tuple[torch.FloatTensor, ...]]],
    ]:
        residual = hidden_states
        hidden_states = self.ln_1(hidden_states)
        attn_outputs = self.attn(
            hidden_states,
            position_ids=position_ids,
            layer_past=layer_past,
            attention_mask=attention_mask,
            head_mask=head_mask,
            use_cache=use_cache,
            output_attentions=output_attentions,
        )
        attn_output = attn_outputs[0]  # output_attn: a, present, (attentions)
        outputs = attn_outputs[1:]
        # residual connection
        hidden_states = attn_output + residual

        if encoder_hidden_states is not None:
            # add one self-attention block for cross-attention
            if not hasattr(self, "crossattention"):
                raise ValueError(
                    f"If `encoder_hidden_states` are passed, {self} has to be instantiated with "
                    "cross-attention layers by setting `config.add_cross_attention=True`"
                )
            residual = hidden_states
            hidden_states = self.ln_cross_attn(hidden_states)
            cross_attn_outputs = self.crossattention(
                hidden_states,
                attention_mask=attention_mask,
                head_mask=head_mask,
                encoder_hidden_states=encoder_hidden_states,
                encoder_attention_mask=encoder_attention_mask,
                output_attentions=output_attentions,
            )
            attn_output = cross_attn_outputs[0]
            # residual connection
            hidden_states = residual + attn_output
            outputs = (
                outputs + cross_attn_outputs[2:]
            )  # add cross attentions if we output attention weights

        residual = hidden_states
        hidden_states = self.ln_2(hidden_states)
        feed_forward_hidden_states = self.mlp(hidden_states)
        # residual connection
        hidden_states = residual + feed_forward_hidden_states

        if use_cache:
            outputs = (hidden_states,) + outputs
        else:
            outputs = (hidden_states,) + outputs[1:]

        return outputs  # hidden_states, present, (attentions, cross_attentions)



import dataclasses
import os
import sys
from typing import Tuple

from cehrbert.runners.hf_runner_argument_dataclass import (
    DataTrainingArguments,
    ModelArguments,
)
from transformers import HfArgumentParser, TrainingArguments
from transformers.utils import logging

from cehrgpt.runners.hf_gpt_runner_argument_dataclass import CehrGPTArguments

LOG = logging.get_logger("transformers")


def parse_dynamic_arguments(
    argument_classes: Tuple[dataclasses.dataclass, ...] = (
        DataTrainingArguments,
        ModelArguments,
        TrainingArguments,
    )
) -> Tuple:
    """
    Parses command-line arguments with extended flexibility, allowing for the inclusion of custom argument classes.

    This function utilizes `HfArgumentParser` to parse arguments from command line input, JSON, or YAML files.
    By default, it expects `ModelArguments`, `DataTrainingArguments`, and `TrainingArguments`, but it can be extended
    with additional argument classes through the `argument_classes` parameter, making it suitable
    for various custom setups.

    Parameters:
        argument_classes (Tuple[Type]): A tuple of argument classes to be parsed. Defaults to
        `(ModelArguments, DataTrainingArguments, TrainingArguments)`. Additional argument classes can be specified
        for greater flexibility in configuration.

    Returns:
        Tuple: A tuple of parsed arguments, one for each argument class provided. The order of the returned tuple
        matches the order of the `argument_classes` parameter.

    Raises:
        FileNotFoundError: If the specified JSON or YAML file does not exist.
        json.JSONDecodeError: If there is an error parsing a JSON file.
        yaml.YAMLError: If there is an error parsing a YAML file.
        Exception: For other issues that occur during argument parsing.

    Example usage:
        - Command-line: `python training_script.py --model_name_or_path bert-base-uncased --do_train`
        - JSON file: `python training_script.py config.json`
        - YAML file: `python training_script.py config.yaml`

    Flexibility:
        The function can be customized to include new argument classes as needed:

        Example with a custom argument class:
            ```python
            class CustomArguments:
                # Define custom arguments here
                pass


            custom_args = parse_extended_args(
                (ModelArguments, DataTrainingArguments, TrainingArguments, CustomArguments)
            )
            ```
        This example demonstrates how to include additional argument classes
        beyond the defaults for a more tailored setup.
    """
    parser = HfArgumentParser(argument_classes)

    # Check if input is a JSON or YAML file
    if len(sys.argv) == 2 and sys.argv[1].endswith(".json"):
        args = parser.parse_json_file(json_file=os.path.expanduser(sys.argv[1]))
    elif len(sys.argv) == 2 and sys.argv[1].endswith(".yaml"):
        args = parser.parse_yaml_file(yaml_file=os.path.expanduser(sys.argv[1]))
    else:
        args = parser.parse_args_into_dataclasses()

    return tuple(args)


def parse_runner_args() -> (
    Tuple[CehrGPTArguments, DataTrainingArguments, ModelArguments, TrainingArguments]
):
    cehrgpt_args, data_args, model_args, training_args = parse_dynamic_arguments(
        (CehrGPTArguments, DataTrainingArguments, ModelArguments, TrainingArguments)
    )
    return cehrgpt_args, data_args, model_args, training_args



import logging
import os
import random
from datetime import datetime

import pandas as pd
from sklearn import metrics

from cehrgpt.models.tokenization_hf_cehrgpt import CehrGptTokenizer

from .utils import (
    RANDOM_SEE,
    create_demographics,
    create_gender_encoder,
    create_race_encoder,
    create_vector_representations,
    find_match,
    scale_age,
)

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)
LOG = logging.getLogger("member_inference")
THRESHOLD = 5


def main(args):
    LOG.info(f"Started loading tokenizer at {args.tokenizer_path}")
    concept_tokenizer = CehrGptTokenizer.from_pretrained(args.tokenizer_path)

    LOG.info(f"Started loading training data at {args.training_data_folder}")
    train_data = pd.read_parquet(args.training_data_folder)

    LOG.info(f"Started loading evaluation data at {args.evaluation_data_folder}")
    evaluation_data = pd.read_parquet(args.evaluation_data_folder)

    LOG.info(f"Started loading synthetic_data at {args.synthetic_data_folder}")
    synthetic_data = pd.read_parquet(args.synthetic_data_folder)

    LOG.info(
        "Started extracting the demographic information from the patient sequences"
    )
    train_data = create_demographics(train_data)
    evaluation_data = create_demographics(evaluation_data)
    synthetic_data = create_demographics(synthetic_data)

    LOG.info("Started rescaling age columns")
    train_data = scale_age(train_data)
    evaluation_data = scale_age(evaluation_data)
    synthetic_data = scale_age(synthetic_data)

    LOG.info("Started encoding gender")
    gender_encoder = create_gender_encoder(train_data, evaluation_data, synthetic_data)
    LOG.info("Completed encoding gender")

    LOG.info("Started encoding race")
    race_encoder = create_race_encoder(train_data, evaluation_data, synthetic_data)
    LOG.info("Completed encoding race")

    random.seed(RANDOM_SEE)
    all_results = []
    for i in range(1, args.n_iterations + 1):
        dist_metrics = []
        LOG.info(f"Iteration {i} Started")
        train_data_sample = train_data.sample(args.num_of_samples)
        evaluation_data_sample = evaluation_data.sample(args.num_of_samples)
        synthetic_data_sample = synthetic_data.sample(args.num_of_samples)
        LOG.info(f"Iteration {i}: Started creating train vectors")
        train_vectors = create_vector_representations(
            train_data_sample, concept_tokenizer, gender_encoder, race_encoder
        )
        LOG.info(f"Iteration {i}: Started creating evaluation vectors")
        evaluation_vectors = create_vector_representations(
            evaluation_data_sample, concept_tokenizer, gender_encoder, race_encoder
        )
        LOG.info(f"Iteration {i}: Started creating synthetic vectors")
        synthetic_vectors = create_vector_representations(
            synthetic_data_sample, concept_tokenizer, gender_encoder, race_encoder
        )
        LOG.info(
            f"Iteration {i}: Started calculating the distances between synthetic and training vectors"
        )
        synthetic_train_dist = find_match(synthetic_vectors, train_vectors)
        synthetic_evaluation_dist = find_match(synthetic_vectors, evaluation_vectors)

        dist_metrics.extend([(_, 1) for _ in synthetic_train_dist])
        dist_metrics.extend([(_, 0) for _ in synthetic_evaluation_dist])

        metrics_pd = pd.DataFrame(dist_metrics, columns=["dist", "label"])
        metrics_pd["pred"] = (metrics_pd.dist < THRESHOLD).astype(int)

        results = {
            "Iteration": i,
            "Accuracy": metrics.accuracy_score(metrics_pd.label, metrics_pd.pred),
            "Precision": metrics.precision_score(metrics_pd.label, metrics_pd.pred),
            "Recall": metrics.recall_score(metrics_pd.label, metrics_pd.pred),
            "F1": metrics.f1_score(metrics_pd.label, metrics_pd.pred),
        }
        all_results.append(results)
        LOG.info(f"Iteration {i}: Privacy loss {results}")

    current_time = datetime.now().strftime("%m-%d-%Y-%H-%M-%S")
    pd.DataFrame(
        all_results, columns=["Iteration", "Accuracy", "Precision", "Recall", "F1"]
    ).to_parquet(
        os.path.join(args.output_folder, f"membership_inference_{current_time}.parquet")
    )


def create_argparser():
    import argparse

    parser = argparse.ArgumentParser(
        description="Membership Inference Analysis Arguments"
    )
    parser.add_argument(
        "--training_data_folder",
        dest="training_data_folder",
        action="store",
        help="The path for where the training data folder",
        required=True,
    )
    parser.add_argument(
        "--evaluation_data_folder",
        dest="evaluation_data_folder",
        action="store",
        help="The path for where the evaluation data folder",
        required=True,
    )
    parser.add_argument(
        "--synthetic_data_folder",
        dest="synthetic_data_folder",
        action="store",
        help="The path for where the synthetic data folder",
        required=True,
    )
    parser.add_argument(
        "--output_folder",
        dest="output_folder",
        action="store",
        help="The output folder that stores the metrics",
        required=True,
    )
    parser.add_argument(
        "--tokenizer_path",
        dest="tokenizer_path",
        action="store",
        help="The path to ConceptTokenizer",
        required=True,
    )
    parser.add_argument(
        "--num_of_samples",
        dest="num_of_samples",
        action="store",
        type=int,
        required=False,
        default=5000,
    )
    parser.add_argument(
        "--n_iterations",
        dest="n_iterations",
        action="store",
        type=int,
        required=False,
        default=1,
    )
    return parser


if __name__ == "__main__":
    main(create_argparser().parse_args())



OBSERVATION_PERIOD_QUERY = """
select
    person_id,
    observation_period_start_date,
    case
        when observation_period_end_date >= add_months(current_date(), -12) then current_date()
        else observation_period_end_date
    end as observation_period_end_date,
    period_type_concept_id
from (
     SELECT person_id,
            MIN(observation_period_start_date) AS observation_period_start_date,
            MAX(observation_period_end_date)   AS observation_period_end_date,
            44814725                           as period_type_concept_id
     FROM (
              SELECT pt.person_id             AS person_id,
                     MIN(vt.visit_start_date) AS observation_period_start_date,
                     MAX(vt.visit_end_date)   AS observation_period_end_date
              FROM person as pt
                       JOIN visit_occurrence as vt ON pt.person_id = vt.person_id
              WHERE YEAR(vt.visit_start_date) >= 1985
                AND vt.visit_start_date <= current_date() --set lower and upper bound to ignore spurious dates
              GROUP BY pt.person_id

              UNION

              SELECT pt.person_id                 AS person_id,
                     MIN(co.condition_start_date) AS observation_period_start_date,
                     MAX(co.condition_start_date) AS observation_period_end_date
              FROM person as pt
                       JOIN condition_occurrence as co ON pt.person_id = co.person_id
              WHERE YEAR(co.condition_start_date) >= 1985
                AND co.condition_start_date <= current_date() --set lower and upper bound to ignore spurious dates
              GROUP BY pt.person_id

              UNION

              SELECT pt.person_id           AS person_id,
                     MIN(po.procedure_date) AS observation_period_start_date,
                     MAX(po.procedure_date) AS observation_period_end_date
              FROM person as pt
                       JOIN procedure_occurrence as po ON pt.person_id = po.person_id
              WHERE YEAR(po.procedure_date) >= 1985
                AND po.procedure_date <= current_date() --set lower and upper bound to ignore spurious dates
              GROUP BY pt.person_id

              UNION

              SELECT pt.person_id                     AS person_id,
                     MIN(de.drug_exposure_start_date) AS observation_period_start_date,
                     MAX(de.drug_exposure_start_date) AS observation_period_end_date
              FROM person as pt
                       JOIN drug_exposure as de ON pt.person_id = de.person_id
              WHERE YEAR(de.drug_exposure_start_date) >= 1985
                AND de.drug_exposure_start_date <= current_date() --set lower and upper bound to ignore spurious dates
              GROUP BY pt.person_id
          ) as x
     WHERE x.observation_period_end_date IS NOT NULL
     GROUP BY person_id
 ) as z
"""

OBSERVATION_PERIOD_WITH_MEASUREMENT_QUERY = """
select
    person_id,
    observation_period_start_date,
    case
        when observation_period_end_date >= add_months(current_date(), -12) then current_date()
        else observation_period_end_date
    end as observation_period_end_date,
    period_type_concept_id
from (
     SELECT person_id,
            MIN(observation_period_start_date) AS observation_period_start_date,
            MAX(observation_period_end_date)   AS observation_period_end_date,
            44814725                           as period_type_concept_id
     FROM (
              SELECT pt.person_id             AS person_id,
                     MIN(vt.visit_start_date) AS observation_period_start_date,
                     MAX(vt.visit_end_date)   AS observation_period_end_date
              FROM person as pt
                       JOIN visit_occurrence as vt ON pt.person_id = vt.person_id
              WHERE YEAR(vt.visit_start_date) >= 1985
                AND vt.visit_start_date <= current_date() --set lower and upper bound to ignore spurious dates
              GROUP BY pt.person_id

              UNION

              SELECT pt.person_id                 AS person_id,
                     MIN(co.condition_start_date) AS observation_period_start_date,
                     MAX(co.condition_start_date) AS observation_period_end_date
              FROM person as pt
                       JOIN condition_occurrence as co ON pt.person_id = co.person_id
              WHERE YEAR(co.condition_start_date) >= 1985
                AND co.condition_start_date <= current_date() --set lower and upper bound to ignore spurious dates
              GROUP BY pt.person_id

              UNION

              SELECT pt.person_id           AS person_id,
                     MIN(po.procedure_date) AS observation_period_start_date,
                     MAX(po.procedure_date) AS observation_period_end_date
              FROM person as pt
                       JOIN procedure_occurrence as po ON pt.person_id = po.person_id
              WHERE YEAR(po.procedure_date) >= 1985
                AND po.procedure_date <= current_date() --set lower and upper bound to ignore spurious dates
              GROUP BY pt.person_id

              UNION

              SELECT pt.person_id                     AS person_id,
                     MIN(de.drug_exposure_start_date) AS observation_period_start_date,
                     MAX(de.drug_exposure_start_date) AS observation_period_end_date
              FROM person as pt
                       JOIN drug_exposure as de ON pt.person_id = de.person_id
              WHERE YEAR(de.drug_exposure_start_date) >= 1985
                AND de.drug_exposure_start_date <= current_date() --set lower and upper bound to ignore spurious dates
              GROUP BY pt.person_id

              UNION

              SELECT
                pt.person_id AS person_id,
                MIN(m.measurement_date) AS observation_period_start_date,
                MAX(m.measurement_date) AS observation_period_end_date
              FROM person as pt
                JOIN measurement as m ON pt.person_id = m.person_id
              WHERE YEAR(m.measurement_date) >= 1985
                AND m.measurement_date <= current_date() --set lower and upper bound to ignore spurious dates
              GROUP BY pt.person_id
          ) as x
     WHERE x.observation_period_end_date IS NOT NULL
     GROUP BY person_id
 ) as z
"""



from cehrgpt.omop.omop_argparse import create_omop_argparse
from cehrgpt.omop.omop_table_builder import OmopTableBuilder
from cehrgpt.omop.queries.condition_era import CONDITION_ERA_QUERY

CONDITION_ERA = "condition_era"


def main(args):
    OmopTableBuilder.create_omop_query_builder(
        input_folder=args.input_folder,
        output_folder=args.output_folder,
        continue_job=args.continue_job,
        table_name=CONDITION_ERA,
        query_template=CONDITION_ERA_QUERY,
        dependency_list=["condition_occurrence"],
    ).build()


if __name__ == "__main__":
    main(create_omop_argparse().parse_args())
